/* GMSD - Group Membership Service Daemon
 * Copyright (C) 2003 Nelio Pereira (nelio@ime.usp.br)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

#include <errno.h>
#include <unistd.h>

/* Trying to set a larger backlog buffer... 
   #include <netinet/in.h>
   #include <sys/types.h>
   #include <sys/socket.h>
*/

/* compilation definitions */
#include "config.h"

//#include <glib.h>
//#include <gnet.h>

#include "comm.h"
#include "log.h"
#include "msg.h"


#include <errno.h>
#include <string.h>


/*! The communication server. */
static comm_server_t *comm_server = NULL; 

/*! Whether this module is enabled or not. */
static gboolean enabled = FALSE;

#if USE_TCP
static void comm_start_server(GServer*, GConn*, gpointer);
#endif

#if USE_TCP || USE_CONN
static void comm_handle_connection(GConn*, GConnEvent*, gpointer);
#endif

static void comm_receive(const gchar* const, const gint);
static gboolean comm_receive_async(GIOChannel*, GIOCondition, comm_state_t* const);


/**
 * Initializes the communication routines and data structures.
 * \param port The port number to listen.
 * \param mcast_address The multicast address.
 * \param mcast_port The multicast port number.
 * \param mcast_ttl The multicast time-to-live value.
 * \param mcast_loop Whether messages sent through multicast will be loopbacked.
 */
void comm_init(const guint port, 
	       const gchar* const mcast_address, const guint mcast_port, const guint mcast_ttl, 
	       const gboolean mcast_loop)
{
  gint rv;
  GIOChannel *channel;

  log_debug("[comm init] init");

  gnet_init();

  /* Create the server */
  comm_server = g_new0(comm_server_t, 1);
#if USE_TCP
  comm_server->tcp = NULL;
#endif
  comm_server->udp = g_new0(comm_state_t, 1);
  comm_server->udp->socket = g_new0(comm_socket_t, 1);
  comm_server->mcast = g_new0(comm_state_t, 1);
  comm_server->mcast->socket = g_new0(comm_socket_t, 1);

  /* TCP server */
#if USE_TCP
  comm_server->tcp = gnet_server_new(NULL, port, comm_start_server, NULL);
  g_assert(comm_server->tcp != NULL);
  log_info("listening to TCP port %d", port);
#endif

  /* UDP server */
  comm_server->udp->socket->type = COMM_SOCKET_UDP;
  comm_server->udp->socket->protocol.udp  = gnet_udp_socket_new_with_port(port);
  g_assert(comm_server->udp->socket->protocol.udp != NULL);
  comm_server->udp->ia = NULL;

  comm_server->udp->watch_flags = COMM_IO_CONDITION;
  channel = gnet_udp_socket_get_io_channel(comm_server->udp->socket->protocol.udp);
  g_assert(channel != NULL);
  comm_server->udp->watch = g_io_add_watch(channel, 
					   comm_server->udp->watch_flags,
					   (GIOFunc)comm_receive_async, 
					   comm_server->udp);
  log_info("listening to UDP port %d", port);
  g_source_set_priority(g_main_context_find_source_by_id(NULL, comm_server->udp->watch),
    			G_PRIORITY_HIGH);

			
  /* MCAST server */
#if USE_MULTICAST
  comm_server->mcast->socket->type = COMM_SOCKET_MCAST;
  comm_server->mcast->socket->protocol.mcast = gnet_mcast_socket_new_with_port(mcast_port);
  g_assert(comm_server->mcast->socket->protocol.mcast != NULL);
  
  /* trying to set a larger backlog buffer...
     GIOChannel *ioc = gnet_mcast_socket_get_io_channel(comm_server->mcast->socket->protocol.mcast);
     gint fd = g_io_channel_unix_get_fd(ioc);
     gint ret = setsockopt(fd, IPPROTO_TCP, SO_RCVBUF, "100", 3);
     g_printf("- %d %s\n", ret, strerror());
     g_assert (ret == 0);
  */

  gnet_mcast_socket_set_loopback(comm_server->mcast->socket->protocol.mcast, mcast_loop);
  gnet_mcast_socket_set_ttl(comm_server->mcast->socket->protocol.mcast, mcast_ttl);

  comm_server->mcast->ia = gnet_inetaddr_new(mcast_address, mcast_port);
  if (comm_server->mcast->ia == NULL) {
    printf("COMM MCAST SERVER ERROR: '%s'\n", strerror(errno));
  }
  g_assert(comm_server->mcast->ia != NULL);

  rv = gnet_mcast_socket_join_group(comm_server->mcast->socket->protocol.mcast, comm_server->mcast->ia);
  g_assert(rv == 0);
  log_info("joined multicast address %s:%d",
	   gnet_inetaddr_get_name(comm_server->mcast->ia),  
	   gnet_inetaddr_get_port(comm_server->mcast->ia));
  comm_server->mcast->watch_flags = COMM_IO_CONDITION;
  channel = gnet_mcast_socket_get_io_channel(comm_server->mcast->socket->protocol.mcast);
  g_assert(channel != NULL);
  comm_server->mcast->watch = g_io_add_watch(channel, 
					     comm_server->mcast->watch_flags,
					     (GIOFunc)comm_receive_async, 
					     comm_server->mcast);
  g_source_set_priority(g_main_context_find_source_by_id(NULL, comm_server->mcast->watch),
    			G_PRIORITY_HIGH);
#endif

  enabled = TRUE;

  log_debug("[comm init] exit");
}



/**
 * Destroys the data structures and finalizes the communication process.
 */
void comm_finalize()
{
  gint rv;

  log_debug("[comm finalize] init");

  /* TCP */
#if USE_TCP
  gnet_server_delete(comm_server->tcp);
  log_info("TCP server stopped");
  g_free(comm_server->tcp);
#endif

  /* UDP */
  g_source_remove(comm_server->udp->watch);
  log_info("UDP server stopped");
  gnet_udp_socket_delete(comm_server->udp->socket->protocol.udp);
  g_free(comm_server->udp->socket);
  g_free(comm_server->udp);

  /* MCAST */
#if USE_MULTICAST
  g_source_remove(comm_server->mcast->watch);
  log_info("MCAST server stopped");
  rv = gnet_mcast_socket_leave_group(comm_server->mcast->socket->protocol.mcast, comm_server->mcast->ia);
  g_assert(rv == 0);
  gnet_mcast_socket_delete(comm_server->mcast->socket->protocol.mcast);
  g_free(comm_server->mcast->socket);
  gnet_inetaddr_delete(comm_server->mcast->ia);
  g_free(comm_server->mcast);
#endif

  g_free(comm_server);
  comm_server = NULL;

  enabled = FALSE;

  log_debug("[comm finalize] exit");
}



/**
 * Indicates whether this module has been initialized.
 * \return \em TRUE if this module has been initialized.
 */
gboolean comm_enabled()
{
  return enabled;
}


#if USE_TCP

/**
 * Represents the TCP server, that receives connections and dispacthes them
 * to be handled by a specific function.
 * \param server The socket server object.
 * \param conn The connection just stablished.
 * \param user_data Specific user data.
 */
static void comm_start_server(GServer* server, GConn* conn, gpointer user_data)
{
  g_assert(server != NULL);
  g_assert(conn != NULL);

  gnet_conn_set_callback (conn, comm_handle_connection, NULL);
  gnet_conn_set_watch_error (conn, TRUE);
  gnet_conn_read(conn);
}

#endif


#if USE_TCP || USE_CONN

/**
 * Handles connections dispacthed by the socket server object.
 * \param conn The connection just stablished.
 * \param event The event occured within this connection. 
 * \param user_data Specific user data.
 */
static void comm_handle_connection(GConn* conn, GConnEvent* event, gpointer user_data)
{
  switch (event->type) {
    
  case GNET_CONN_READ:
    comm_receive(event->buffer, event->length);
    gnet_conn_read(conn);
    break;
    
  case GNET_CONN_WRITE:
    gnet_conn_write(conn, event->buffer, event->length);
    break;
    
  case GNET_CONN_CLOSE:
  case GNET_CONN_TIMEOUT:
  case GNET_CONN_ERROR:
    log_notice("[comm] connection closed from host %s:%d", conn->hostname, conn->port);
    gnet_conn_delete (conn);
    break;
    
  default:
    g_assert_not_reached();

  }
}

#endif



/**
 * Receives an arbitrary message.
 * \param buffer The buffer where the message is stored.
 * \param length The length of the buffer.
 */
static void comm_receive(const gchar* const buffer, const gint length)
{
  msg_t *msg;
  xmlDocPtr doc;

  log_debug("[comm receive] init");

  doc = xmlParseMemory(buffer, length);
  msg = msg_from_xml(doc);
  g_assert(msg != NULL);
  xmlFreeDoc(doc);
  msg_receive(msg);
  msg_destroy_full(msg);

  log_debug("[comm receive] exit");
}



/**
 * Routine called to asynchronously receive messages.
 * \param iochannel The GIOChannel by which the message has arrived.
 * \param condition The condition signaled on the channel. 
 * \param comm_state The communication state of the channel.
 * \return \em TRUE if the read has been succesfully completed.
 */
static gboolean comm_receive_async(GIOChannel *iochannel, GIOCondition condition, comm_state_t* comm_state)
{
  gboolean rv = TRUE;

  g_assert(comm_state != NULL);

  log_debug("[comm receive async] init");

  /* Check for socket error */
  if (condition & (G_IO_ERR | G_IO_HUP | G_IO_NVAL)) {
    log_notice("error: socket error (%d, %d)\n", condition, errno);
    rv = FALSE;
  }

  /* Check for data to be read */
  else if (condition & G_IO_IN) {
    GIOStatus read_status;
    comm_buffer_t buffer;
    GError *error = NULL;

    /* Read the data into our buffer */
    read_status = g_io_channel_read_chars(iochannel, 
					  buffer.data, 
					  COMM_BUFFER_SIZE, 
					  &buffer.size, 
					  &error);
    if ((read_status == G_IO_STATUS_ERROR) || (read_status == G_IO_STATUS_AGAIN)) {
      log_info("[comm receive async] error receiving message: %s\n", error->message);
      rv = FALSE;
    } 
#if 0
    else {
      guint old_watch_flags;
      g_assert (bytes_read > 0);
      
      comm_state->buffer = g_renew(gchar, comm_state->buffer, comm_state->size + bytes_read);
      g_snprintf(comm_state->buffer + comm_state->size, bytes_read, buffer);
      comm_state->size += bytes_read;
      printf("BUF %d: '%s'\n", comm_state->size, comm_state->buffer);

      old_watch_flags = comm_state->watch_flags;
      
      /* Remove the IN watch if the data is over */
      if (buffer[bytes_read-2] == '\0') { 
	comm_state->watch_flags &= ~G_IO_IN;
      	printf("Lido! '%s'\n", comm_state->buffer); 
	comm_receive(comm_state->buffer, comm_state->size);
	g_free(comm_state->buffer);
      }
      else {
      /* Update watch flags if they changed */
	//      if (old_watch_flags != comm_state->watch_flags) {
	printf("source remove\n");
	g_source_remove(comm_state->watch);
	comm_state->watch = g_io_add_watch(iochannel, 
					    comm_state->watch_flags, 
					    (GIOFunc)comm_receive_async, 
					    comm_state);
	//	g_source_set_priority(g_main_context_find_source_by_id(NULL, comm_comm_state->watch),
	//			      G_PRIORITY_HIGH);
	rv = FALSE;
      }
    }
#else
    comm_receive(buffer.data, buffer.size);
#endif
  }
  
  log_debug("[comm receive async] exit");
  
  return rv;
}



/**
 * Sends a piece of data to a destination over TCP.
 * \param hostname The host to send the the message.
 * \param port The port where the host is waiting for the message.
 * \param data The data to be sent.
 * \param length The length of the buffer where the data is stored.
 */
void comm_send_tcp(const gchar* const hostname, const guint port, const gchar* const data, const gsize length)
{
  GTcpSocket *socket;

  log_debug("[comm send tcp] init");

#if USE_CONN
  GConn *conn = gnet_conn_new(hostname, port, comm_handle_connection, NULL);
  gnet_conn_connect(conn);
  gnet_conn_write(conn, data, length);
  gnet_conn_disconnect(conn);
  gnet_conn_delete(conn);
#else
  /* Create the socket */
  socket = gnet_tcp_socket_connect(hostname, port);
  if (socket == NULL) {
    /* could not connect to host. ignore */
    log_debug("[comm send tcp] could not connect to %s:%d", hostname, port);
  } else {
    gint bytes_written = 0;
    GIOChannel *iochannel;
    gchar *buffer;
    GIOError ioerror = G_IO_ERROR_NONE;
    GError *error = NULL;

    /* Get the IOChannel */
    iochannel = gnet_tcp_socket_get_io_channel(socket);
    g_assert (iochannel != NULL);
    g_io_channel_set_buffered (iochannel, TRUE);
    g_io_channel_set_encoding (iochannel, NULL, &error);
    
    buffer = g_new0(gchar, length+1);
    g_snprintf(buffer, length+1, "%s%c", data, '\0');
    ioerror = gnet_io_channel_writen(iochannel, buffer, length+1, &bytes_written);
    if (error != G_IO_ERROR_NONE) {
      log_info("error %d while writing into iochannel. %d bytes written.", ioerror, bytes_written);
    }
    g_free(buffer);
    log_debug("[comm send tcp] message sent to %s:%d", hostname, port);
    
    gnet_tcp_socket_delete(socket);
  }
#endif

  log_debug("[comm send tcp] exit");
} 



/**
 * Sends a piece of data to multiple destinations over TCP.
 * \param n The number of hosts to send the data.
 * \param hostname The host names to send the data.
 * \param port The port ports on which the hosts are waiting the message.
 * \param data The data to be sent.
 * \param length The length of the buffer where the data is stored.
 */
void comm_send_tcp_n(const guint n, gchar** const hostname, guint* const port, const gchar* const data, const gsize length)
{
  guint i;

  log_debug("[comm send tcp n] init");

  for (i = 0; i < n; i++) {
    comm_send_tcp(hostname[i], port[i], data, length);
  }

  log_debug("[comm send tcp n] exit");
}



/**
 * Sends a piece of data to a destination over UDP.
 * \param hostname The host to send the the message.
 * \param port The port where the host is waiting for the message.
 * \param data The data to be sent.
 * \param length The length of the buffer where the data is stored.
 */
void comm_send_udp(const gchar* const hostname, const guint port, const gchar* const data, const gsize length)
{
  GUdpSocket *socket;
  GInetAddr *dest;
  gint rv;

  log_debug("[comm send udp] init");

  socket = gnet_udp_socket_new();
  g_assert(socket != NULL);
  dest = gnet_inetaddr_new(hostname, port);
  if (dest == NULL) {
    log_notice("could not lookup address %s:%d", hostname, port);
  } else {
    g_assert(dest != NULL);
    rv = gnet_udp_socket_send(socket, data, length, dest);
    g_assert(rv == 0);
    log_debug("[comm send udp] message sent to %s:%d", hostname, port);
    gnet_inetaddr_delete(dest);
    gnet_udp_socket_delete(socket);
  }

  log_debug("[comm send udp] exit");
}



/**
 * Sends a piece of data to multple destinations over UDP.
 * \param n The number of hosts to send the data.
 * \param hostname The host names to send the data.
 * \param port The port ports on which the hosts are waiting the message.
 * \param data The data to be sent.
 * \param length The length of the buffer where the data is stored.
 */
void comm_send_udp_n(const guint n, gchar** const hostname, guint* const port, const gchar* const data, const gsize length)
{
  guint i;

  log_debug("[comm send udp n] init");

  for (i = 0; i < n; i++) {
    comm_send_udp(hostname[i], port[i], data, length);
  }

  log_debug("[comm send udp n] exit");
}



/**
 * Sends information through a multicast channel.
 * \param data The data to be sent.
 * \param length The length of the data.
 */
#if USE_MULTICAST
void comm_send_mcast(const gchar* const data, const gsize length)
{
  gint rv;
  gchar *buffer;
  /* sleep time between send retries */
  const guint retry_timeout = 1;

  log_debug("[comm send mcast] init");

  /* if this is not valid, the other node wouldn't be able to receive this data, */
  g_assert(length+1 <= COMM_BUFFER_SIZE);

  buffer = g_new0(gchar, length+1);
  g_snprintf(buffer, length+1, "%s%c", data, '\0');
  
  rv = gnet_mcast_socket_send(comm_server->mcast->socket->protocol.mcast, 
			      buffer, 
			      length+1, 
			      comm_server->mcast->ia);

  /* retry until message is sent. probably this is not a very good approach. */
  while (rv != 0) {
    log_warning("Failed to send message to %s:%d - retrying in %d seconds", 
		gnet_inetaddr_get_name(comm_server->mcast->ia),  
		gnet_inetaddr_get_port(comm_server->mcast->ia),
		retry_timeout);
    sleep(retry_timeout);
    rv = gnet_mcast_socket_send(comm_server->mcast->socket->protocol.mcast, 
				buffer, 
				length+1, 
				comm_server->mcast->ia);
  }
  //  g_assert(rv == 0);  

  log_debug("[comm send mcast] message sent to %s:%d", 
  	    gnet_inetaddr_get_name(comm_server->mcast->ia),  
  	    gnet_inetaddr_get_port(comm_server->mcast->ia));
  
  g_free(buffer);

  log_debug("[comm send mcast] exit");
}
#endif
