/* GMSD - Group Membership Service Daemon
 * Copyright (C) 2003 Nelio Pereira (nelio@ime.usp.br)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

#ifndef _COMM_H
#define _COMM_H

#include <glib.h>
#include <gnet.h>


/*! The default communication port number. */
#define COMM_PORT_DEFAULT 5555

/*! The default multicast address. */
#define COMM_MCAST_ADDRESS_DEFAULT "sap.mcast.net"

/*! The default multicast port number. */
#define COMM_MCAST_PORT_DEFAULT 5555

/*! The default multicast TTL. */
#define COMM_MCAST_TTL_DEFAULT 1

/*! The default multicast loopback setup. */
#define COMM_MCAST_LOOP_DEFAULT TRUE

/*! The watch flags used in IO channels in this module. */
#define COMM_IO_CONDITION G_IO_IN | G_IO_PRI |G_IO_HUP | G_IO_NVAL

/*! The maximum buffer size. This may cause a limit on the number of members of the group!! */
#define COMM_BUFFER_SIZE 131072


/*! The socket types definition. */
typedef enum {
  COMM_SOCKET_TCP = 1,   /*!< A TCP socket. */
  COMM_SOCKET_UDP = 2,   /*!< An UDP socket. */
  COMM_SOCKET_MCAST = 3  /*!< A multicast socket. */
} comm_socket_type_t;


/*! The socket type. */
typedef struct {
  union {
    GTcpSocket *tcp;         /*!< The TCP socket. */
    GUdpSocket *udp;         /*!< The UDP socket. */
    GMcastSocket *mcast;     /*!< The multicast socket. */
  } protocol;                /*!< The communication protocol used by this socket. */
  comm_socket_type_t type;   /*!< The socket type. */
} comm_socket_t;


/*! The communication buffer definition. */
typedef struct {
  gchar data[COMM_BUFFER_SIZE];  /*!< The input data. */
  gint size;                     /*!< The buffer size. */
} comm_buffer_t;


/*! The communication state definition. */
typedef struct {
  comm_socket_t *socket;        /*!< The socket. */
  GInetAddr *ia;                /*!< The Inet Address. */
  guint watch_flags;            /*!< The flags used to wait for. */
  guint watch;                  /*!< The event identifier in the main_loop. */
} comm_state_t;


/*! The communication server definition. */
typedef struct {
  GServer *tcp;         /*!< The TCP server. */
  comm_state_t *udp;    /*!< The UDP server. */
  comm_state_t *mcast;  /*!< The MCAST server. */
} comm_server_t;


void comm_init(const guint, const gchar* const, const guint, const guint, const gboolean);
void comm_finalize(void);
gboolean comm_enabled(void);

void comm_send_tcp(const gchar* const, const guint, const gchar* const, const gsize);
void comm_send_tcp_n(const guint,gchar** const, guint* const, const gchar* const, const gsize);
void comm_send_udp(const gchar* const, const guint, const gchar* const, const gsize);
void comm_send_udp_n(const guint, gchar** const, guint* const, const gchar* const, const gsize);
void comm_send_mcast(const gchar* const, const gsize);

#endif /* comm.h */


