/* GMSD - Group Membership Service Daemon
 * Copyright (C) 2003 Nelio Pereira (nelio@ime.usp.br)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

/* compilation definitions */
#include "config.h"

#include <errno.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include <confuse.h>
//#include <glib.h>

#include "comm.h"
#include "configuration.h"
#include "gmsd.h"
#include "ipc.h" 
#include "log.h"
#include "protocol.h"
#include "timer.h"


/*! The configuration file name.*/
gchar *config_filename = NULL;

/*! The configuration parsed from \ref config_filename. */
cfg_t *configuration = NULL;

static gboolean config_parse(void);
static int config_verify_node_file(cfg_t*, cfg_opt_t*, const char*, void*);
static int config_verify_log_level(cfg_t*, cfg_opt_t*, const char*, void*);


/*! The options that will be held on the configuration file.  */
static cfg_opt_t config_opts[] = {
  CFG_INT(CONFIG_OPT_PORT, COMM_PORT_DEFAULT, CFGF_NONE),
  CFG_STR(CONFIG_OPT_MCAST_ADDRESS, COMM_MCAST_ADDRESS_DEFAULT, CFGF_NONE),
  CFG_INT(CONFIG_OPT_MCAST_PORT, COMM_MCAST_PORT_DEFAULT, CFGF_NONE),
  CFG_BOOL(CONFIG_OPT_MCAST_LOOP, COMM_MCAST_LOOP_DEFAULT, CFGF_NONE),
  CFG_INT(CONFIG_OPT_MCAST_TTL, COMM_MCAST_TTL_DEFAULT, CFGF_NONE),
  CFG_STR_CB(CONFIG_OPT_NODE_LIST_FILE, NODE_LIST_FILE_DEFAULT, CFGF_NONE, &config_verify_node_file),
  //CFG_STR(CONFIG_OPT_NODE_LIST_FILE, NODE_LIST_FILE_DEFAULT, CFGF_NONE),
  CFG_INT(CONFIG_OPT_TIMEOUT_DELAY, PROTOCOL_TIMEOUT_DELAY_DEFAULT, CFGF_NONE),
  CFG_INT(CONFIG_OPT_TIMEOUT_NEIGHBOR, PROTOCOL_TIMEOUT_NEIGHBOR_DEFAULT, CFGF_NONE),
  CFG_INT(CONFIG_OPT_TIMEOUT_PROBE, PROTOCOL_TIMEOUT_PROBE_DEFAULT, CFGF_NONE),
  CFG_STR(CONFIG_OPT_LOG_FILENAME, LOG_FILENAME_DEFAULT, CFGF_NONE),
  CFG_INT_CB(CONFIG_OPT_LOG_LEVEL, LOG_LEVEL_INFO_VALUE, CFGF_NONE, &config_verify_log_level),
  CFG_STR(CONFIG_OPT_GROUP_ID_FILENAME, GROUP_ID_FILENAME_DEFAULT, CFGF_NONE),
  CFG_STR(CONFIG_OPT_IPC_FILENAME, IPC_FILENAME_DEFAULT, CFGF_NONE),
  CFG_STR(CONFIG_OPT_RUN_AS_USER, RUN_AS_USER, CFGF_NONE),
  CFG_STR(CONFIG_OPT_RUN_AS_GROUP, RUN_AS_GROUP, CFGF_NONE),
  CFG_END()
};
  


/**
 * Initializes everytinh related to the configuration file.
 * \param file The configuration file that should be read.
 */
void config_init(gchar* const file) 
{
  log_debug("[config init] init");

  if (file == NULL) {
    guint size;
    size = strlen(CONFIG_FILENAME_DEFAULT) + 1;
    config_filename = g_new0(gchar, size);
    g_snprintf(config_filename, size, "%s", CONFIG_FILENAME_DEFAULT);
  } else {
    config_filename = g_strdup(file);
  }

  /* ensure that path name is absolute. */
  if (! g_path_is_absolute(config_filename)) {
    config_filename = g_strjoin("/", g_get_current_dir(), config_filename, NULL); 
  }

  config_read();

  log_debug("[config init] exit");
}


/**
 * Destroys all structures used with config parsing.
 */
void config_finalize() 
{
  log_debug("[config finalize] init");

  cfg_free(configuration);
  configuration = NULL;
  g_free(config_filename);
  config_filename = NULL;

  log_debug("[config finalize] exit");
}


/**
 * Informs if this modules has been initialized.
 * \return A boolean that indicates if this modules has been initialized.
 */
gboolean config_enabled()
{
  return ((configuration != NULL) && (config_filename != NULL));
}


/**
 * Reads the configuration file.
 */
void config_read() 
{
  if (! config_parse()) {
    log_error("could not read configuration file");
    exit(EX_NOINPUT);
  }
}


/**
 * Parse the configuration file, and fills \ref config_opts.
 * \return Indicates the success or failure in parsing the configuration file.
 */
static gboolean config_parse() 
{
  gint c;

  cfg_free(configuration);
  configuration = cfg_init(config_opts, CFGF_NONE);

  c = cfg_parse(configuration, config_filename);
  switch (c) {
    
  case CFG_SUCCESS:
    return TRUE;
    break;
    
  case CFG_FILE_ERROR:
    log_error("error reading configuration file %s: %s", config_filename, g_strerror(errno));
    perror(config_filename);
    return FALSE;
    break;
    
  case CFG_PARSE_ERROR:
    log_error("error parsing configuration file %s", config_filename);
    return FALSE;
    break;

  default:
    g_assert_not_reached();    
    log_error("undefined return from cfg_parse: %d", c);
    return FALSE;
  }
}


/**
 *
 * \param cfg
 * \param opt
 * \param value
 * \param result
 */
int config_verify_node_file(cfg_t* cfg, cfg_opt_t* opt, const char *value, void *result)
{
  if (! access(value, R_OK)) {
    *(char **)result = g_strdup(value);
    return 0;
  } else {
    cfg_error(cfg, "Invalid value for option %s (%s): %s", opt->name, value, g_strerror(errno));
    return -1;
  }
}


/**
 *
 * \param cfg
 * \param opt
 * \param value
 * \param result
 */
static int config_verify_log_level(cfg_t * cfg, cfg_opt_t * opt, const char *value, void *result)
{
  if (strcmp(value, LOG_LEVEL_ERROR) == 0)
    *(long int *) result = LOG_LEVEL_ERROR_VALUE;
  else if (strcmp(value, LOG_LEVEL_CRITICAL) == 0)
    *(long int *) result = LOG_LEVEL_CRITICAL_VALUE;
  else if (strcmp(value, LOG_LEVEL_WARNING) == 0)
    *(long int *) result = LOG_LEVEL_WARNING_VALUE;
  else if (strcmp(value, LOG_LEVEL_NOTICE) == 0)
    *(long int *) result = LOG_LEVEL_NOTICE_VALUE;
  else if (strcmp(value, LOG_LEVEL_INFO) == 0)
    *(long int *) result = LOG_LEVEL_INFO_VALUE;
  else if (strcmp(value, LOG_LEVEL_DEBUG) == 0)
    *(long int *) result = LOG_LEVEL_DEBUG_VALUE;
  else {
    cfg_error(cfg, "Invalid value for option %s: %s", opt->name, value);
    return -1;
  }
  return 0;
}
