/* GMSD - Group Membership Service Daemon
 * Copyright (C) 2003 Nelio Pereira (nelio@ime.usp.br)
 *
 * This application is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This application is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

/* compilation definitions */
#include "config.h"

#include <errno.h>
#include <sys/mman.h>
#include <sysexits.h>
#include <unistd.h>

#include <grp.h>
#include <pwd.h>
#include <sys/types.h>


#include <confuse.h>
//#include <glib.h>

#include "cmdline.h"
#include "comm.h"
#include "configuration.h"
#include "gmsd.h"
#include "ipc.h"
#include "log.h"
#include "protocol.h"
#include "signals.h"
//#include "timer.h" 
#include "util.h"
#include "xml.h"


/*! The main task of the program. */
GMainLoop *mainloop;


/*! The command line arguments structure. */
extern cmdline_t *cmdline_args;
/*! The configuration options structure. */
extern cfg_t     *configuration;


/*
 * Initialize the configuration parameters (command line and 
 * configuration file).
 * \param argc
 * \param argv
 */
void gmsd_init_config(const guint argc, gchar** const argv) 
{
  cmdline_init();
  if (! cmdline_parse(argc, argv)) {
    cmdline_usage(argv[0]);
    exit(EX_USAGE);
  }

  config_init(cmdline_args->config_filename);
}


/**
 * Initializes the membership daemon routines and modules.
 * \param argc
 * \param argv
 */
void gmsd_init()
{
  node_t* node; 
  gchar* nodename;
  guint length;

  /* let's initialize first the membership module to load the nodes
     configuration, because the node identifiers are defined there.
  */
  membership_init(cfg_getstr(configuration, CONFIG_OPT_NODE_LIST_FILE),
		  cfg_getint(configuration, CONFIG_OPT_PORT));

  /* the node name to be used in the log will be just the node identifier. */
  node = membership_get_mynode();
  length = util_num_length(node->id) + 1;
  nodename = g_new0(gchar, length);
  g_snprintf(nodename, length, NODE_ID_FORMAT, node->id);

  /* let's init now the log, passing the node name to be used. */
  log_init(cmdline_args->debug ? LOG_LEVEL_DEBUG_VALUE : cfg_getint(configuration, CONFIG_OPT_LOG_LEVEL), 
	   nodename,
	   cfg_getstr(configuration, CONFIG_OPT_LOG_FILENAME)
	   );
  g_free(nodename);
  
  log_debug("[gmsd init] init");
  log_info("starting gmsd server...");

  mainloop = g_main_loop_new(NULL, FALSE);

  xml_init();
  signals_init();
  comm_init(cfg_getint(configuration, CONFIG_OPT_PORT),
	    cfg_getstr(configuration, CONFIG_OPT_MCAST_ADDRESS),
	    cfg_getint(configuration, CONFIG_OPT_MCAST_PORT),
	    cfg_getint(configuration, CONFIG_OPT_MCAST_TTL),
	    cfg_getbool(configuration, CONFIG_OPT_MCAST_LOOP));
  ipc_init(cfg_getstr(configuration, CONFIG_OPT_IPC_FILENAME));

  msg_init(membership_get_mynode());

  /* finally, let's start the protocol */
  /*
  protocol_init(cfg_getstr(configuration, CONFIG_OPT_GROUP_ID_FILENAME),
		cfg_getint(configuration, CONFIG_OPT_TIMEOUT_NEIGHBOR),
  		cfg_getint(configuration, CONFIG_OPT_TIMEOUT_PROBE),
  		cfg_getint(configuration, CONFIG_OPT_TIMEOUT_DELAY));
  */

  log_info("initialization process terminated without errors."); 
  log_debug("[gmsd init] exit");
}


/**
 * Frees all memory and destroys all structures used in this program.
 */
void gmsd_finalize()
{
  log_debug("[gmsd finalize] init");

  
  if (protocol_enabled())   protocol_finalize();
  if (msg_enabled())        msg_finalize();
  if (membership_enabled()) membership_finalize();
  if (ipc_enabled())        ipc_finalize(); 
  if (comm_enabled())       comm_finalize();
  if (signals_enabled())    signals_finalize();
  if (xml_enabled())        xml_finalize();
  if (config_enabled())     config_finalize();
  if (cmdline_enabled())    cmdline_finalize();
  

  g_main_loop_unref(mainloop);

  log_info("server stopped.");
  log_debug("[gmsd finalize] exit");

  if (log_enabled())        log_finalize();
}


/**
 * Stops the main loop. 
 */
void gmsd_stop()
{
  log_debug("[gmsd stop] init");

  g_main_loop_quit(mainloop);

  log_debug("[gmsd stop] exit");
}


/**
 * Runs this program in \em daemon mode.
 */
static void gmsd_daemonize() 
{
  switch(fork()) {
    
  case 0:  /* child process */
    break;
    
  case -1: /* fork failed */
    exit(EX_OSERR);
    break;

  default: /* parent process */
    exit(EX_OK);
    break;
  }
  
  setsid();
  chdir("/");
  
  /* Do not close open files */
}


gint main(const int argc, gchar** const argv) 
{
  /* 
     First just initialize command line arguments 
     and configuration file to know if I have to fork. 
  */
  gmsd_init_config(argc, argv);

  if (cmdline_args->daemon && ! cmdline_args->debug) {
    gmsd_daemonize();
  }

  gmsd_init();

  /* Now I can lock correctly my memory. */
#if HAVE_MLOCKALL

  /* if run as root, become another user */
  if (geteuid() == 0) {
    struct passwd *pw;
    struct group  *gw;
    gchar *user  = cfg_getstr(configuration, CONFIG_OPT_RUN_AS_USER);
    gchar *group = cfg_getstr(configuration, CONFIG_OPT_RUN_AS_GROUP);
    
    log_info("setting uid:gid to %s:%s", user, group);

    if (mlockall(MCL_CURRENT | MCL_FUTURE) != 0) {
      log_notice("error locking memory: %s", g_strerror(errno));
    }

    pw = getpwnam(user);
    g_assert(pw != NULL);

    gw = getgrnam(group);
    g_assert(gw != NULL);
    
    setregid(gw->gr_gid, gw->gr_gid);
    setreuid(pw->pw_uid, pw->pw_uid);

  } else {
    log_notice("cannot lock memory: only root can do that.");
  }

#endif

  log_info("server up. waiting for events."); 

  /* finally, let's start the protocol */
  protocol_init(cfg_getstr(configuration, CONFIG_OPT_GROUP_ID_FILENAME),
		cfg_getint(configuration, CONFIG_OPT_TIMEOUT_NEIGHBOR),
  		cfg_getint(configuration, CONFIG_OPT_TIMEOUT_PROBE),
  		cfg_getint(configuration, CONFIG_OPT_TIMEOUT_DELAY));

  g_main_loop_run(mainloop);
  log_info("server down. releasing resources.");

#if HAVE_MLOCKALL
  /**
   * As I became a regular user, I cannot unlock my memory. However,
   * the process is going to be over anyway soon..
   *
   
   if (run_as_root) {
   seteuid(0);
   setegid(0);
   
   if (munlockall() != 0) {
   log_notice("error unlocking memory: %s", g_strerror(errno));
   }
   }
   
   *
   ***/
#endif
  
  gmsd_finalize();

  return EX_OK;
}
