/* GMSD - Group Membership Service Daemon
 * Copyright (C) 2003 Nelio Pereira (nelio@ime.usp.br)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

#include <string.h>

//#include <glib.h>
#include "glib2_compat.h"

#include "group.h"
#include "log.h"
//#include "node.h"
#include "util.h"
#include "xml.h"



/**
 *
 * \return 
 */
group_t *group_new() 
{
  group_t *group = g_new0(group_t, 1);
  group->members = NULL;
  group->size = 0;
  return group;
}


/**
 *
 * \param group
 */
void group_destroy(group_t* group)
{
  //  log_debug("[group destroy] init");

  g_list_free(group->members);
  g_free(group);

  //  log_debug("[group destroy] exit");
}


/**
 *
 * \param group
 */
void group_destroy_full(group_t* group)
{
  //  log_debug("[group destroy_full] init");

  g_assert(group != NULL);
  g_list_foreach(group->members, (GFunc)node_destroy, NULL);
  group_destroy(group);

  //  log_debug("[group destroy_full] exit");
}


/**
 *
 * \param group
 */
guint group_size(const group_t* const group) 
{
  //  return g_list_length(group->members);
  return group->size;
}


/**
 *
 * \param group
 * \param node
 */ 
void group_add_node(group_t* const group, node_t* const node) 
{
  g_assert(group != NULL);
  group->members = g_list_insert_sorted(group->members, node, (GCompareFunc)node_compare);
  group->size++;
}


/**
 *
 * \param group
 * \param node
 */ 
void group_remove_node(group_t* const group, const node_t* const node) 
{
  GList *p;

  g_assert(group != NULL);
  p = g_list_find_custom(group->members, node, (GCompareFunc)node_compare);
  g_assert(p != NULL);

  group->members = g_list_delete_link(group->members, p);
  group->size--;
}


/**
 *
 * \param group
 * \param node
 * \return 
 */
gboolean group_in(const group_t* const group, const node_t* const node)
{
  return (group_find(group, node) != NULL);
}


/**
 * Finds the node equal to \em node in the given group.
 * \param group The group to look for the node.
 * \param node The node to look for in the group.
 * \return A pointer to the node in the group, or NULL if the node doesn't exist in the group.
 */
node_t* group_find(const group_t* const group, const node_t* const node)
{
  GList *l;
  
  l = g_list_find_custom(group->members, node, (GCompareFunc)node_compare);
  return (l == NULL) ? NULL : (node_t*)l->data;
}



/**
 * Creates a string that represents a given group.
 * \param group The group to be stringfied.
 * \return A string that represents the group.
 */
gchar* group_to_string(const group_t* const group)
{
  gchar *str = NULL;
  GList *list;
  node_t *p;
  gchar **str_array = NULL;
  guint i, size, length;
  
  size = group_size(group);
  str_array = g_new0(gchar*, size+1);
  for (i = 0, list = group->members; list != NULL; i++, list = g_list_next(list)) {
    p = (node_t*)list->data;
    length = util_num_length(p->id) + 1;
    str_array[i] = g_new0(gchar, length);
    g_snprintf(str_array[i], length, NODE_ID_FORMAT, p->id);
  }
  str_array[size] = NULL;

  str = g_strjoinv(",", str_array);

  g_strfreev(str_array);

  return str;
}


/**
 * Prints the elements of a group.
 * \param group The group to enumerate.
 */
void group_print(const group_t* const group)
{
  GList *list;
  node_t *p;
  
  g_assert(group != NULL);
  G_PRINTF("+ - - - - - - - - - - - - - - - - - - - -\n");
  G_PRINTF("| Group size: %d\n", group_size(group));
  for (list = group->members; list != NULL; list = g_list_next(list)) {
    p = (node_t*)list->data;
    G_PRINTF("| %d -> %s:%d\n", p->id, p->hostname, p->port);
  }
  G_PRINTF("+ - - - - - - - - - - - - - - - - - - - -\n");
}


/**
 *
 * \param group
 * \return xmlDocPtr
 */
xmlDocPtr group_to_xml(const group_t* const group)
{
  xmlDocPtr doc, doc_node;
  xmlNodePtr members;
  GList *list;
  xmlChar *group_size;
  guint length;
  
  //  log_debug("[group to xml] init");

  doc = xmlNewDoc(XML_VERS);
  doc->children = xmlNewDocRawNode(doc, NULL, XML_GROUP, NULL);

  length = util_num_length(group->size) + 1;
  group_size = g_new(xmlChar, length);
  g_snprintf(group_size, length, "%u", group->size);
  xmlNewChild(doc->children, NULL, XML_GROUP_SIZE, group_size);
  g_free(group_size);

  members = xmlNewChild(doc->children, NULL, XML_GROUP_MEMBERS, NULL);
  for (list = group->members; list != NULL; list = list->next) {
    doc_node = node_to_xml((node_t*)list->data);
    xmlAddChild(members, xmlCopyNode(doc_node->children, 1));
    xmlFreeDoc(doc_node);
  }

  //  log_debug("[group to xml] exit");

  return doc;
}


/**
 *
 * \param doc
 * \return
 */
group_t* group_from_xml(const xmlDocPtr doc)
{
  group_t *group;

  //  log_debug("[group from xml] init");

  if (strcmp(doc->children->name, XML_GROUP)) {
    /* not a group xml!!! */
    group = NULL;
  } else {
    guint group_size;
    xmlDocPtr doc_aux;
    xmlNodePtr cur;
    xmlChar *aux;
    guint i;
    node_t *node;
    
    group = group_new();
    
    cur = doc->children->children;
    aux = xmlNodeGetContent(cur);
    group_size = atoi(aux);
    xmlFree(aux);
    
    for (i = 0, cur = cur->next->children; i < group_size; i++, cur = cur->next) {
      doc_aux = xmlNewDoc(XML_VERS);
      doc_aux->children = xmlCopyNode(cur, 1);
      node = node_from_xml(doc_aux);
      g_assert(node != NULL);
      xmlFreeDoc(doc_aux);

      group_add_node(group, node);
    }
  }

  //  log_debug("[group from xml] exit");

  return group;
}

