/* GMSD - Group Membership Service Daemon
 * Copyright (C) 2003 Nelio Pereira (nelio@ime.usp.br)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

#include <errno.h>
#include <string.h>
#include <unistd.h>

//#include <glib.h>

#include "group_id.h"
#include "log.h"
#include "util.h"
#include "xml.h"


/*! */
static FILE *group_id_fp = NULL;


/**
 * Creates a new group identifier.
 * \param seq The sequence number of the group identifier.
 * \param nodeid The node identifier of the group identifier.
 * \return A new group identifier.
 */
group_id_t* group_id_new(const group_id_seq_t seq, const node_id_t nodeid)
{
  group_id_t *gid = g_new0(group_id_t, 1);
  group_id_set(gid, seq, nodeid);
  return gid;
}


/**
 *
 * \param gid
 * \param filename
 */
void group_id_persist_init(group_id_t* gid, const gchar* const filename)
{
  gboolean exist;

  g_assert(gid != NULL);
  
  exist = g_file_test(filename, G_FILE_TEST_IS_REGULAR | G_FILE_TEST_EXISTS);
  
  group_id_fp = fopen(filename, "a+");
  if (group_id_fp == NULL) {
    log_error("error opening gid file '%s': %s", filename, g_strerror(errno));
  }
  rewind(group_id_fp);
  
  if (exist) {
    group_id_load(gid);
  } else {
    group_id_store(gid);
  }
}


/**
 *
 */
void group_id_persist_destroy()
{
  fclose(group_id_fp);
  group_id_fp = NULL;
}

/**
 * Creates a copy of a given group identifier.
 * \param group_id The group identifier to be copied.
 * \return A copy of a group identifier.
 */
group_id_t* group_id_copy(const group_id_t* const group_id)
{
  return group_id_new(group_id->seq, group_id->node_id);
}

/**
 *
 * \return 
 */
gboolean group_id_persist_enabled()
{
  return (group_id_fp != NULL);
}


/**
 *
 * \param gid
 */
void group_id_destroy(group_id_t *gid)
{
  g_free(gid);
}


/**
 *
 * \param gid
 */
void group_id_store(group_id_t* const gid)
{
  ftruncate(fileno(group_id_fp), 0);
  fprintf(group_id_fp, GROUP_ID_SEQ_FORMAT, gid->seq);
  rewind(group_id_fp);
}


/**
 *
 * \param gid
 */
void group_id_load(group_id_t* const gid)
{
  group_id_seq_t seq;
  fscanf(group_id_fp, GROUP_ID_SEQ_FORMAT, &seq);
  rewind(group_id_fp);

  group_id_set(gid, seq, gid->node_id);
}


/**
 *
 * \param gid
 * \param seq
 * \param nodeid
 */
void group_id_set(group_id_t* const gid, const group_id_seq_t seq, const node_id_t nodeid)
{
  gid->seq     = seq;
  gid->node_id = nodeid;
}


/**
 * \brief Compares two group identifiers.
 *
 * The function should return a negative integer if the first group identifier comes before the second, 
 * 0 if they are equal, or a positive integer if the first group identifier comes after the second. 
 * The comparison follows the definition:
 * \f[ g_1 < g_2 \Longleftrightarrow (g_1.seq < g_2.seq) \vee (g_1.seq = g_2.seq \wedge g_1.node\_id < g_2.node\_id) \f]
 * \param group_id1 First gid. 
 * \param group_id2 Second gid.
 * \return negative value if group_id1 < group_id2; zero if group_id1 = group_id2; positive value if group_id1 > group_id2.
 */ 
gint group_id_compare(const group_id_t* const group_id1, 
		      const group_id_t* const group_id2) 
{
  gint x;

  g_assert(group_id1 != NULL);
  g_assert(group_id2 != NULL);

  return (x = group_id1->seq - group_id2->seq) ? x : node_id_compare(group_id1->node_id, group_id2->node_id);
}


/**
 * Converts a group identifier to a XML structure.
 * \param group_id The group identifier to be converted.
 * \return A XML created from the group identifier.
 */
xmlDocPtr group_id_to_xml(const group_id_t* const group_id)
{
  xmlDocPtr doc;
  guint length;
  xmlChar *aux;

  //  log_debug("[group_id to xml] init");

  doc = xmlNewDoc(XML_VERS);
  doc->children = xmlNewDocRawNode(doc, NULL, XML_GROUP_ID, NULL);

  length = util_num_length(group_id->seq) + 1;
  aux = g_new0(xmlChar, length);
  g_snprintf(aux, length, GROUP_ID_SEQ_FORMAT, group_id->seq);
  xmlSetProp(doc->children, XML_GROUP_ID_SEQUENCE, aux);
  g_free(aux);

  length = util_num_length(group_id->node_id) + 1;
  aux = g_new0(xmlChar, length);
  g_snprintf(aux, length, NODE_ID_FORMAT, group_id->node_id);
  xmlSetProp(doc->children, XML_GROUP_ID_NODE_ID, aux);
  g_free(aux);

  //  log_debug("[group_id to xml] exit");

  return doc;
}


/**
 *
 * \param doc
 * \return
 */
group_id_t* group_id_from_xml(const xmlDocPtr doc)
{
  group_id_t* group_id;

  //  log_debug("[group_id from xml] init");

  if (strcmp(doc->children->name, XML_GROUP_ID)) {
    /* not a group_id xml!!! */
    group_id = NULL;
  } else {
    group_id_seq_t seq;
    node_id_t node_id;
    xmlChar *aux;

    aux = xmlGetProp(doc->children, XML_GROUP_ID_SEQUENCE);
    seq = atoi(aux);
    xmlFree(aux);
    
    aux = xmlGetProp(doc->children, XML_GROUP_ID_NODE_ID);
    node_id = atoi(aux);
    xmlFree(aux);
    
    group_id = group_id_new(seq, node_id);
  }
  
  //  log_debug("[group_id from xml] exit");

  return group_id;
}
