/* GMSD - Group Membership Service Daemon
 * Copyright (C) 2003 Nelio Pereira (nelio@ime.usp.br)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <time.h>

//#include <glib.h>
#include "glib2_compat.h"

#include "gmsd.h"
#include "log.h"
#include "util.h" 


/*! The number of lines that will be buffered in memory before being logged. */
#define LOG_BUFFER_SIZE 50


/*! The log. */ 
static log_t *log = NULL;

/*! The line counter used to flush the log into a file. */
static guint line_count = 0;

/*! The log buffer. */
static GString *log_buffer = NULL;


static void log_handler(const gchar*, GLogLevelFlags, const gchar*, log_t* const);
static gchar* log_get_level(GLogLevelFlags);
static void log_flush(const log_t* const, GString *);



/**
 * Initializes the logging routines.
 * \param level
 * \param nodename
 * \param file
 */
void log_init(const gint level, const gchar* const nodename, const gchar* const file) 
{
  GLogLevelFlags flags = 1 << level;
  log = log_new(LOG_TAG,
		nodename,
		flags | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, 
		file,
		(GLogFunc)log_handler);
  log_buffer = g_string_new(NULL);
}


/**
 * Destroy the log handling routines and frees all memory used.
 */
void log_finalize() 
{
  log_destroy(log);

  log = NULL;
}


/**
 * Informs if this modules has been initialized.
 * \return A boolean that indicates if this modules has been initialized.
 */
gboolean log_enabled()
{
  return (log != NULL);
}


/**
 * Creates a new log.
 * \param domain 
 * \param nodename
 * \param level
 * \param file
 * \param func
 * \returns A new log.
 */
log_t* log_new(const gchar* const domain, 
	       const gchar* const nodename,
	       GLogLevelFlags level, 
	       const gchar* const file, 
	       GLogFunc func)
{
  GError *error = NULL;
  log_t* log = g_new0(log_t, 1);
  log->domain = domain;
  log->nodename = g_strdup(nodename);
  log->level = level;
  log->handler = g_log_set_handler(domain, LOG_ALL_LEVEL, func, log);
  log->iochannel = g_io_channel_new_file(file, "a", &error);
  if (error != NULL) {
    G_FPRINTF(stderr, error->message);
    g_assert_not_reached();
  }
  return log;
}


/**
 * Destroys a log.
 * \param log The log to be destroyed.
 */
void log_destroy(log_t* log)
{
  GError *error = NULL;

  /* deleting the log buffer */
  log_flush(log, log_buffer);
  g_string_free(log_buffer, TRUE);

  g_free(log->nodename);
  g_log_remove_handler(log->domain, log->handler);
  g_io_channel_shutdown(log->iochannel, TRUE, &error);
  g_assert(error == NULL);

  g_free(log);
}


/**
 *
 * \param domain
 * \param level
 * \param message
 * \param log
 */
static void log_handler(const gchar *domain, 
			GLogLevelFlags level, 
			const gchar* message, 
			log_t* log)
{
  if (log->level >= level) { 
    gchar *buffer, *level_str;
    gchar date[LOG_DATE_LENGTH];
    guint length;
    GTimeVal gtv;
    
    g_get_current_time(&gtv);
    ctime_r(&gtv.tv_sec, date);
    date[strlen(date)-1] = '\0';
    level_str = log_get_level(level);

    length = strlen(date) + 6 + strlen(domain) + strlen(log->nodename) + 
      strlen(level_str) + strlen(message) + 14; 
    
    buffer = g_new0(gchar, length);
    g_snprintf(buffer, length, "[%s - %6.6ld] %s %s <%s>: %s\n", 
    	       date, gtv.tv_usec, domain, log->nodename, level_str, message);
    g_string_append(log_buffer, buffer);

    /* is it time to flush the buffer? */
    line_count = (line_count + 1) % LOG_BUFFER_SIZE; 
    if (line_count == 0) {
      log_flush(log, log_buffer);
    }  

    if (log->level & G_LOG_LEVEL_DEBUG) {
      G_FPRINTF(stdout, buffer);
    }

    g_free(buffer);
  }
}


/**
 * 
 * \param log
 * \param log_buffer
 */
static void log_flush(const log_t* const log, GString *log_buffer)
{
  GError *error = NULL;
  
  if (g_io_channel_write_chars(log->iochannel, log_buffer->str, log_buffer->len, NULL, &error) != G_IO_STATUS_NORMAL) {
    G_FPRINTF(stderr, error->message);
  }

  error = NULL;
  g_io_channel_flush(log->iochannel, &error);
  g_assert(error == NULL);

  //g_string_free(log_buffer, TRUE);
  g_string_truncate(log_buffer, 0);
}


/**
 * Exits immediatelly freeing resources.
 */
static void log_exit()
{
  g_log(LOG_TAG, G_LOG_LEVEL_CRITICAL, "exiting");
  gmsd_finalize();
  exit(EX_OSERR);
}


/**
 * Returns a string representing a log level.
 * \param log_level A log level.
 * \return A string that represents the log level.
 */
static gchar* log_get_level(GLogLevelFlags log_level)
{
  if (log_level & G_LOG_LEVEL_ERROR) {
    return LOG_LEVEL_ERROR;
  } 
  if (log_level & G_LOG_LEVEL_CRITICAL) {
    return LOG_LEVEL_CRITICAL;
  }
  if (log_level & G_LOG_LEVEL_WARNING) {
    return LOG_LEVEL_WARNING;
  }
  if (log_level & G_LOG_LEVEL_MESSAGE) {
    return LOG_LEVEL_NOTICE;
  }
  if (log_level & G_LOG_LEVEL_INFO) {
    return LOG_LEVEL_INFO;
  }
  if (log_level & G_LOG_LEVEL_DEBUG) {
    return LOG_LEVEL_DEBUG;
  }
  return NULL;
}


/**
 * Log error messages. 
 * \param format 
 * \param ... 
 */
void log_error(const char* const format, ...) 
{
  va_list args;

  if (! log_enabled()) return;
  va_start(args, format);
  g_logv(LOG_TAG, G_LOG_LEVEL_ERROR, format, args);
  va_end(args);
  log_exit();
}


/**
 * Log critical messages. 
 * \param format 
 * \param ... 
 */
void log_critical(const char* const format, ...) 
{
  va_list args;

  if (! log_enabled()) return;
  va_start(args, format);
  g_logv(LOG_TAG, G_LOG_LEVEL_CRITICAL, format, args);
  va_end(args);
  log_exit();
}


/**
 * Log warning messages. 
 * \param format 
 * \param ... 
 */
void log_warning(const char* const format, ...) 
{
  va_list args;

  if (! log_enabled()) return;
  va_start(args, format);
  g_logv(LOG_TAG, G_LOG_LEVEL_WARNING, format, args);
  va_end(args);
}


/**
 * Log notice messages. 
 * \param format 
 * \param ... 
 */
void log_notice(const char* const format, ...) 
{
  va_list args;

  if (! log_enabled()) return;
  va_start(args, format);
  g_logv(LOG_TAG, G_LOG_LEVEL_MESSAGE, format, args);
  va_end(args);
}


/**
 * Log information messages. 
 * \param format 
 * \param ... 
 */
void log_info(const char* const format, ...) 
{
  va_list args;

  if (! log_enabled()) return;
  va_start(args, format);
  g_logv(LOG_TAG, G_LOG_LEVEL_INFO, format, args);
  va_end(args);
}


/**
 * Log debug messages. 
 * \param format 
 * \param ... 
 */
void log_debug(const char* const format, ...) 
{
  va_list args;

  if (! log_enabled()) return;
  va_start(args, format);
  g_logv(LOG_TAG, G_LOG_LEVEL_DEBUG, format, args);
  va_end(args);
}



/**
 * 
 * \return
 */
gboolean log_info_enabled()
{
  return (log->level >= G_LOG_LEVEL_INFO);
}


/**
 * 
 * \return
 */
gboolean log_notice_enabled()
{
  return (log->level >= G_LOG_LEVEL_MESSAGE);
}


/**
 * 
 * \return
 */
gboolean log_warning_enabled()
{
  return (log->level >= G_LOG_LEVEL_WARNING);
}


/**
 * 
 * \return
 */
gboolean log_critical_enabled()
{
  return (log->level >= G_LOG_LEVEL_CRITICAL);
}


/**
 * 
 * \return
 */
gboolean log_error_enabled()
{
  return (log->level >= G_LOG_LEVEL_ERROR);
}


/**
 * 
 * \return
 */
gboolean log_debug_enabled()
{
  return (log->level >= G_LOG_LEVEL_DEBUG);
}

