/* GMSD - Group Membership Service Daemon
 * Copyright (C) 2003 Nelio Pereira (nelio@ime.usp.br)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

#include <string.h>
#include <unistd.h>

//#include <glib.h>
#include "glib2_compat.h"

//#include "group.h"
//#include "group_id.h"
#include "log.h"
#include "membership.h"
//#include "node.h"
#include "util.h"
#include "xml.h"


/*! The membership list of all possible nodes in the group. */
group_t* membership_list = NULL;

/*! The current membership. */
membership_t* membership = NULL;

/*! This node. */
static node_t* membership_mynode = NULL;



static void membership_load_members(const gchar* const, const guint);
//static void membership_print_timestamps(gpointer, gpointer, gpointer);
static void membership_to_xml_timestamps(gpointer, gpointer, xmlNodePtr);



/**
 * Initializes the membership handling routines.
 * \param nodes_file_name The file where are located all possible nodes in the membership.
 * \param port The port which this node will be listening.
 */
void membership_init(const gchar* const nodes_file_name, const guint port) 
{
  log_debug("[membership init] init");

  membership_list = group_new();
  membership_load_members(nodes_file_name, port);
  g_assert(membership_mynode != NULL);

  membership = membership_new(group_id_new(GROUP_ID_INITIAL_ID, membership_mynode->id));

  log_debug("[membership init] exit");
}


/**
 * Destroys all structures related to membership.
 */
void membership_finalize()
{  
  log_debug("[membership finalize] init");
  
  /* 
     The call below deletes all nodes in the membership list.
     Notice that all nodes in the program are references to a
     node in this list.
  */
  group_destroy_full(membership_list);
  
  /* 
     we shouldn't call membership_destroy_full on this membership, 
     because it's already partially destroyed.
  */
  group_destroy(membership->group);
  membership_destroy(membership);
  
  /* not necessary because of the first statement. */
  //node_destroy(membership_mynode);
  
  log_debug("[membership finalize] exit");
}


/**
 * Informs if this modules has been initialized.
 * \return A boolean that indicates if this modules has been initialized.
 */
gboolean membership_enabled()
{
  return ((membership_list != NULL) && 
	  (membership != NULL) &&
	  (membership_mynode != NULL));
}


/**
 * Creates a new membership with a given group identifier.
 * \param gid The group identifier to use to create a new membership.
 * \return A new membership.
 */
membership_t* membership_new(group_id_t* const gid)
{
  membership_t* membership = g_new(membership_t, 1);
  membership->gid = gid;
  membership->group = group_new();
  membership->timestamps = g_hash_table_new(NULL, NULL);
  return membership;
}


/**
 * Destroys a given membership.
 * \param membership The membership to be destroyed.
 */
void membership_destroy(membership_t* membership)
{
  //  log_debug("[membership destroy] init");

  g_hash_table_destroy(membership->timestamps);
  group_id_destroy(membership->gid);
  g_free(membership);

  //  log_debug("[membership destroy] exit");
}


/**
 * Destroys a given membership and all related structures.
 * \param membership The membership to be destroyed.
 */
void membership_destroy_full(membership_t* membership)
{
  //  log_debug("[membership destroy_full] init");

  group_destroy_full(membership->group);
  membership_destroy(membership);

  //  log_debug("[membership destroy_full] exit");
}



/**
 * Returns the local node.
 * \return The local node, or \em NULL if the node hasn't been initialized.
 */
node_t* membership_get_mynode()
{
  return membership_mynode;
}


/**
 * Prints to stdout a given membership.
 * \param membership The membership to be printed.
 */
void membership_print(const membership_t* const membership)
{
  G_PRINTF("+ - - - - - - - - - - - - - - - - - - - -\n");
  G_PRINTF("|         MEMBERSHIP: " NODE_ID_FORMAT "\n", membership_mynode->id);
  G_PRINTF("| - gid: (" GROUP_ID_SEQ_FORMAT "," NODE_ID_FORMAT ")\n", 
	   membership->gid->seq, membership->gid->node_id);
  G_PRINTF("| - group:\n");
  group_print(membership->group);
  //  G_PRINTF("| - timestamps:\n");
  //  g_hash_table_foreach(membership->timestamps, membership_print_timestamps, NULL);
  //  G_PRINTF("+ - - - - - - - - - - - - - - - - - - - -\n");
}




/**
 * Prints the timestamp of a node.
 * \param key The node identifier.
 * \param value The timestamp of the node.
 * \param user_data (unused)
 */
//static void membership_print_timestamps(gpointer key, gpointer value, gpointer user_data)
//{
//  G_PRINTF("| " NODE_ID_FORMAT " -> %d\n", GPOINTER_TO_INT(key), GPOINTER_TO_INT(value));
//}



/**
 * Load all possible membership nodes from the configuration file.
 * \param nodes_file_name
 * \param server_port
 */
static void membership_load_members(const gchar* const nodes_file_name, const guint server_port)
{
  gchar hostname[NODE_NAME_MAX_LENGTH];
  gchar localhost[NODE_NAME_MAX_LENGTH];
  guint port;
  node_t* p;
  gchar *line;
  gsize length;
  GIOChannel *channel;
  GError *error = NULL;
  GIOStatus iostatus;

  guint node_id = 1;
  
  gethostname(localhost, NODE_NAME_MAX_LENGTH);

  g_clear_error(&error);
  channel = g_io_channel_new_file(nodes_file_name, "r", &error);
  g_assert(channel != NULL);

  g_clear_error(&error);
  while ((iostatus = g_io_channel_read_line(channel, &line, &length, NULL, &error)) != G_IO_STATUS_EOF) {
    if (! G_STR_HAS_PREFIX(line, "#") && (sscanf(line, "%s\t%d\n", hostname, &port) == 2)) {
      p = node_new(node_id++, hostname, port);
      if (! strcmp(hostname, localhost) && (port == server_port)) {
	membership_mynode = p;
      }
      group_add_node(membership_list, p);
    }
    g_free(line);
    g_clear_error(&error);
  }

  g_clear_error(&error);
  iostatus = g_io_channel_shutdown(channel, TRUE, &error);
  g_assert(iostatus == G_IO_STATUS_NORMAL);
  
  if (membership_mynode == NULL) {
    log_error("local host (%s:%d) not found in node list (%s)", localhost, server_port, nodes_file_name);
  } 
}


/**
 * Gets a node by its node identifier.
 * \param node_id The node identifier to be used.
 * \return The node for the given node_id;
 */
node_t* membership_get_node_by_id(const node_id_t node_id)
{
  GList *q;
  node_t *p;

  g_assert(membership_list != NULL);

  /* creates a dummy node, only useful for its node_id field. */
  p = node_new(node_id, NULL, 0);

  q = g_list_find_custom(membership_list->members, p, (GCompareFunc)node_compare_by_id);
  node_destroy(p);

  return (q != NULL) ? (node_t*)q->data : NULL;
}


/**
 * Determines if a membership has the majority of members.
 * \param m The membership to be verified.
 * \return TRUE if the membership \em m contains a majority of nodes in 
 *         P. This predicate is assumed to satisfy the requirement:
 *         \f[ majority(m) \wedge majority(m') \Rightarrow m \cap m' \neq \emptyset \f]
 */
gboolean membership_majority(const membership_t* const m) 
{
  g_assert(m != NULL);
  g_assert(m->group != NULL);
  g_assert(membership_list != NULL);

  return group_size(m->group) >= (group_size(membership_list)/2);
}


/**
 *
 * \param m
 * \return The processor with the lowest id in membership m. NULL will be 
 *         returned if the member set is empty.
 */
node_t* membership_leader(const membership_t* const m) 
{
  GList *p;
  g_assert(m != NULL);
  g_assert(m->group != NULL);

  p = g_list_first(m->group->members);
  return (p != NULL) ? (node_t*)p->data : NULL;
}


/**
 *
 * \param m
 * \param p
 * \return The left neighbor or processor \em p in membership \em m.
 */
node_t* membership_left_neighbor(const membership_t* const m, const node_t* const p)
{
  GList *node_p, *q;
  g_assert(p != NULL);
  g_assert(m != NULL);
  g_assert(m->group != NULL);

  node_p = g_list_find_custom(m->group->members, p, (GCompareFunc)node_compare);
  g_assert(node_p != NULL);

  q = g_list_previous(node_p);
  q = (q != NULL) ? q : g_list_last(m->group->members);

  return (node_t*)q->data;
}


/**
 * 
 * \param m
 * \param p
 * \return The right neighbor or processor \em p in membership \em m.
 */
node_t* membership_right_neighbor(const membership_t* const m, const node_t* const p)
{
  GList *node_p, *q;
  g_assert(p != NULL);
  g_assert(m != NULL);
  g_assert(m->group != NULL);

  node_p = g_list_find_custom(m->group->members, p, (GCompareFunc)node_compare);
  g_assert(node_p != NULL);

  q = g_list_next(node_p);
  q = (q != NULL) ? q : g_list_first(m->group->members);

  return (node_t*)q->data;
}


/**
 * Calculates the rank of a given node in a membership.
 * \param m The membership.
 * \param p The node to calculate the rank.
 * \return An integer between 1 and |m| indicating the position of \em p
 *         in the cyclic neighbor relation; the leader has the highest rank.
 */
gint membership_rank(const membership_t* const m, const node_t* const p)
{
  gint index;

  g_assert(p != NULL);
  g_assert(m != NULL);
  g_assert(m->group != NULL);

  index = g_list_index(m->group->members, p);
  g_assert(index != -1);

  return (index == 0) ? group_size(m->group) : index;
}


/**
 * Verifies if the given node is the leader of the given membership.
 * \param m The membership.
 * \param p The node to be verified.
 * \return \em TRUE if the node is the leader of the membership; \em FALSE otherwise.
 */
gboolean membership_is_leader(const membership_t* const m, const node_t* const p)
{
  node_t *leader = membership_leader(membership);
  return (node_compare(leader, membership_mynode) == 0);
}


/**
 * Saves a membership on a file.
 * \param membership The membership to be saved
 * \param filename The file to write the membership.
 */
void membership_save(const membership_t* const membership, const gchar* filename)
{
  xmlDocPtr doc = membership_to_xml(membership);
  xmlSetDocCompressMode(doc, 0);
  xmlSaveFormatFile(filename, doc, 1);
  xmlFreeDoc(doc);
}


/*
  
<?xml version="1.0"?>
<membership>
   <group_id sequence="93" node_id="1"/>
   <group>
      <size>1</size>
      <members>
         <node id="1">nalvesp.ath.cx</node>
      </members>
   </group>
   <timestamps>
      <node id="1">1053787194</node>
      <node id="2">0</node>
   </timestamps>
</membership>

*/

/**
 *
 * \param membership
 * \return xmlDocPtr
 */
xmlDocPtr membership_to_xml(const membership_t* const membership)
{
  xmlDocPtr doc, doc_aux;
  xmlNodePtr cur;
  
  //  log_debug("[membership to xml] init");

  doc = xmlNewDoc(XML_VERS);
  doc->children = xmlNewDocRawNode(doc, NULL, XML_MEMBERSHIP, NULL);
  cur = doc->children;

  doc_aux = group_id_to_xml(membership->gid);
  xmlAddChild(cur, xmlCopyNode(doc_aux->children, 1));
  xmlFreeDoc(doc_aux);

  doc_aux = group_to_xml(membership->group);
  xmlAddChild(cur, xmlCopyNode(doc_aux->children, 1));
  xmlFreeDoc(doc_aux);

  cur = xmlNewChild(doc->children, NULL, XML_MEMBERSHIP_TIMESTAMPS, NULL);
  g_hash_table_foreach(membership->timestamps, (GHFunc)membership_to_xml_timestamps, cur);

  //  log_debug("[membership to xml] exit");

  return doc;
}


/**
 *
 * \param key
 * \param value
 * \param nodePtr
 */
static void membership_to_xml_timestamps(gpointer key, gpointer value, xmlNodePtr nodePtr)
{
  gint node_id = GPOINTER_TO_INT(key);
  glong timestamp = GPOINTER_TO_INT(value);
  xmlNodePtr cur;
  xmlChar *aux;
  guint length;
  
  length = util_num_length(timestamp) + 1;
  aux = g_new(xmlChar, length);
  g_snprintf(aux, length, MEMBERSHIP_TIMESTAMP_FORMAT, timestamp);
  cur = xmlNewChild(nodePtr, NULL, XML_MEMBERSHIP_TIMESTAMPS_NODE, aux);
  g_free(aux);

  length = util_num_length(node_id) + 1;
  aux = g_new(xmlChar, length);
  g_snprintf(aux, length, NODE_ID_FORMAT, node_id);
  xmlSetProp(cur, XML_MEMBERSHIP_TIMESTAMPS_NODE_ID, aux);
  g_free(aux);
}


/**
 *
 * \param doc
 * \return
 */
membership_t* membership_from_xml(const xmlDocPtr doc)
{
  membership_t *memb;

  //  log_debug("[membership from xml] init");

  if (strcmp(doc->children->name, XML_MEMBERSHIP)) {
    /* not a membership xml!!! */
    memb = NULL;
    //    return NULL;
  } else {
    xmlDocPtr doc_aux;
    xmlNodePtr cur;
    xmlChar *aux;
    node_id_t node_id;
    time_t timestamp;
    group_id_t *group_id;
    group_t *group;
    GHashTable *timestamps;
    
    memb = g_new(membership_t, 1);
    
    doc_aux = xmlNewDoc(XML_VERS);
    
    cur = doc->children->children;
    doc_aux->children = cur;
    group_id = group_id_from_xml(doc_aux);
    g_assert(group_id != NULL);
    memb->gid = group_id;
    
    cur = cur->next;
    doc_aux->children = cur;
    group = group_from_xml(doc_aux);
    g_assert(group != NULL);
    memb->group = group;
    
    doc_aux->children = NULL;
    xmlFreeDoc(doc_aux);
    
    cur = cur->next->children;
    timestamps = g_hash_table_new(NULL, NULL);
    for (; cur != NULL; cur = cur->next) {
      aux = xmlGetProp(cur, XML_MEMBERSHIP_TIMESTAMPS_NODE_ID);
      node_id = atoi(aux);
      xmlFree(aux);
      aux = xmlNodeGetContent(cur);
      timestamp = atoi(aux);
      xmlFree(aux);
      g_hash_table_insert(timestamps, GINT_TO_POINTER(node_id), GINT_TO_POINTER(timestamp));
    }
    memb->timestamps = timestamps;
  }

  //  log_debug("[membership from xml] init");
  
  return memb;
}
