/* GMSD - Group Membership Service Daemon
 * Copyright (C) 2003 Nelio Pereira (nelio@ime.usp.br)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

#include <string.h>

//#include <glib.h>
#include "glib2_compat.h"
#include <gnet.h>

#include "log.h"
#include "msg.h"
#include "comm.h"
#include "util.h"
#include "xml.h"

#if USE_MSG_ID
#include <math.h>
#endif


/*! The hashtable that stores the functions to be called to each message type. */
static GHashTable* msg_receive_funcs = NULL;

/*! This node. */
static node_t* msg_mynode;


#if USE_MSG_ID
static guint msg_cur_id = 0;
#endif


/*! The current message being processed. */
static GQueue* processing_messages;


/**
 * Creates a new message.
 * \param from The node sender.
 * \param rcpt The recipient of the message. Might be a node or a group.
 * \param type The type of the message.
 * \param info The information carried on this message.
 * \return A new message.
 */
msg_t* msg_new(node_t* const from, msg_rcpt_t* const rcpt, const msg_type_t type, const xmlDocPtr info)
{
  msg_t *msg = g_new0(msg_t, 1);
  msg->header = g_new0(msg_header_t, 1);
  msg->header->from = from;
  msg->header->rcpt = rcpt;
  msg->header->type = type;
  msg->data = g_new0(msg_data_t, 1);
  msg->data->info = xmlCopyDoc(info, 1);
  
#if USE_MSG_ID
  {
    guint id = ++msg_cur_id;
    guint length = util_num_length(msg_mynode->id) + 1 + util_num_length(id) + 1;
    msg->id = g_new0(gchar, length);
    g_snprintf(msg->id, length, NODE_ID_FORMAT ".%ud", msg_mynode->id, id);
  }
#endif

  return msg;
}


/**
 * Destroys a message.
 * \param msg The message to be destroyed.
 */
void msg_destroy(msg_t* msg)
{
  g_free(msg->header);
  xmlFreeDoc(msg->data->info);
  g_free(msg->data);
#if USE_MSG_ID
  g_free(msg->id);
#endif
  g_free(msg);
}



/**
 * Destroys a message and all related structures.
 * \param msg The message to be destroyed.
 */
void msg_destroy_full(msg_t* msg)
{
  if (msg->header->rcpt->type == MSG_RCPT_NODE) {
    node_destroy(msg->header->rcpt->to.node);
  } else if (msg->header->rcpt->type == MSG_RCPT_GROUP) {
    group_destroy_full(msg->header->rcpt->to.group);
  }
  g_free(msg->header->rcpt);
  node_destroy(msg->header->from);
  msg_destroy(msg);
}


/**
 * Init all message handling structures.
 */
void msg_init(node_t* const mynode)
{
  msg_receive_funcs = g_hash_table_new(NULL, NULL);
  processing_messages = g_queue_new();
  msg_mynode = mynode;
}



/**
 * Finalizes all messages handling structures.
 */
void msg_finalize()
{
  g_hash_table_destroy(msg_receive_funcs);
  g_queue_free(processing_messages);
  msg_receive_funcs = NULL;
  msg_mynode = NULL;
}



/**
 * Indicates whether this module has been initialized.
 * \return \em TURE if this module is initialized.
 */
gboolean msg_enabled()
{
  return msg_receive_funcs != NULL;
}



/**
 * Prints a message to the standard output.
 * \param msg The messsage to be printed.
 */
void msg_print(const msg_t* const msg)
{
  G_PRINTF("+ - - - - - - - - - - - - - - - - - - -\n");
  G_PRINTF("|              MESSAGE\n");
#if USE_MSG_ID
  G_PRINTF("| ID: %s\n", msg->id);
#endif
  G_PRINTF("| HEADER:\n");
  G_PRINTF("|   - from: (%d,%s,%d)\n", msg->header->from->id, 
	   msg->header->from->hostname, msg->header->from->port); 
  G_PRINTF("|   - to:\n");
  switch (msg->header->rcpt->type) {
    
  case MSG_RCPT_NODE:
    break;
    
  case MSG_RCPT_GROUP:
    group_print(msg->header->rcpt->to.group);
    break;

  default:
    G_PRINTF("|      unknown recipient type: %d\n", msg->header->rcpt->type);
  }
  G_PRINTF("|   - type: %d\n|\n", msg->header->type);
  G_PRINTF("| DATA: ");
  xmlSaveFile("-", msg->data->info);
  G_PRINTF("+ - - - - - - - - - - - - - - - - - - - -\n");
}



/**
 * Sends a message to its recipient.
 * \param msg The message to be sent.
 */
void msg_send(msg_t* const msg)
{
  xmlChar *buffer;
  gint size;
  xmlDocPtr doc;

  log_debug("[msg send] init");

  g_queue_push_head(processing_messages, msg);

  g_assert(msg != NULL);

#if USE_MSG_ID
  log_notice("sending message " MSG_ID_TYPE_FORMAT " (type " MSG_TYPE_FORMAT ", recipient type " MSG_RCPT_TYPE_FORMAT ").", 
 	     msg->id, msg->header->type, msg->header->rcpt->type);
#endif

  doc = msg_to_xml(msg);
  xmlDocDumpMemory(doc, &buffer, &size);
  xmlFreeDoc(doc);

  if (msg->header->rcpt->type == MSG_RCPT_NODE) {
    node_t *node = msg->header->rcpt->to.node;
    comm_send_udp(node->hostname, node->port, buffer, size);
  } else if (msg->header->rcpt->type == MSG_RCPT_GROUP) {
#if USE_MULTICAST
    comm_send_mcast(buffer, size);
#else 
    node_t* node;
    gchar** hostname;
    guint* port;
    GList *list;
    guint i, length;
    
    length = group_size(msg->header->rcpt->to.group);
    hostname = g_new0(gchar*, length);
    port = g_new0(guint, length);
    for (list = msg->header->rcpt->to.group->members, i=0; list != NULL; list = g_list_next(list), i++) {
      node = (node_t*)list->data;
      hostname[i] = node->hostname;
      port[i] = node->port;
    }
    comm_send_udp_n(i, hostname, port, buffer, size);
    g_free(hostname);
    g_free(port);
#endif
  }
  xmlFree(buffer);

#if USE_MSG_ID
  log_notice("message " MSG_ID_TYPE_FORMAT " sent (type " MSG_TYPE_FORMAT ", recipient type " MSG_RCPT_TYPE_FORMAT ").", 
 	     msg->id, msg->header->type, msg->header->rcpt->type);
#endif

  g_queue_pop_head(processing_messages);

  log_debug("[msg send] exit");
}



/**
 * Called when a new message has arrived.
 * \param msg The message received.
 */
void msg_receive(msg_t* const msg)
{
  gboolean for_me = FALSE;

  log_debug("[msg receive] init");

  g_assert(msg != NULL);
  g_assert(msg->header != NULL);

  g_queue_push_head(processing_messages, msg);

#if USE_MSG_ID
  log_notice("message " MSG_ID_TYPE_FORMAT " received (type " MSG_TYPE_FORMAT ", recipient type " MSG_RCPT_TYPE_FORMAT ").",
	     msg->id, msg->header->type, msg->header->rcpt->type);
#endif

  if (msg->header->rcpt->type == MSG_RCPT_NODE) {
    if (node_compare(msg->header->rcpt->to.node, msg_mynode) == 0) {
      for_me = TRUE;
    }
  } else if (msg->header->rcpt->type == MSG_RCPT_GROUP) {
    if (group_in(msg->header->rcpt->to.group, msg_mynode)) {
      for_me = TRUE;
    }
  } else {
    log_notice("unknown recipient type %d", msg->header->rcpt->type);
  }

  /* only process the message if it was addressed to me */
  if (for_me) {
    msg_receive_func_t func = (msg_receive_func_t)g_hash_table_lookup(msg_receive_funcs, 
								      GINT_TO_POINTER(msg->header->type));
    if (func == NULL) {
      log_error("message type %d not assigned to any handling routines", msg->header->type);
    }
    func(msg);
  } else {
#if USE_MSG_ID
    log_notice("message " MSG_ID_TYPE_FORMAT " not addressed to me. dropped.", msg->id);
#else
    log_notice("message not addressed to me. dropped.");
#endif
  }

  g_queue_pop_head(processing_messages);
  
  log_debug("[msg receive] exit");
}


/**
 * Returns the current message being processed.
 * \return The current message being processed.
 */
msg_t* msg_get_current_message() 
{
  return (msg_t*)g_queue_peek_head(processing_messages);
}


/**
 * Adds a callback function a a given message type.
 * \param msg_type The message type to assign the callback function.
 * \param receive_func The callback function.
 */
void msg_add_handle(const msg_type_t msg_type, const msg_receive_func_t receive_func)
{
  g_hash_table_insert(msg_receive_funcs, GINT_TO_POINTER(msg_type), receive_func);
}



/**
 * Removes a callback function to a given message type.
 * \param msg_type The message type.
 */
void msg_remove_handle(const msg_type_t msg_type)
{
  g_hash_table_remove(msg_receive_funcs, GINT_TO_POINTER(msg_type));
}



/*
    
  <message>
     <header>
        <from><node id="">...</node></from>
	<to><node id="">...</node></to>
	<type>123</type>
     </header>
     <data>
        ...
     </data>
  </message>

  */

/**
 * Converts a message into an XML structure.
 * \param msg The message to be converted.
 * \return A document created from a message. This document should be freed when no longer needed.
 */
xmlDocPtr msg_to_xml(const msg_t* const msg) 
{
  xmlDocPtr doc, doc_aux;
  xmlNodePtr cur, aux;
  xmlChar *type;
  guint length;

  //  log_debug("[msg to xml] init");

  doc = xmlNewDoc(XML_VERS);
  doc->children = xmlNewDocRawNode(doc, NULL, XML_MESSAGE, NULL);
#if USE_MSG_ID
 {
   xmlChar *aux;
   aux = g_strdup(msg->id);
   xmlSetProp(doc->children, XML_MESSAGE_ID, aux);
   g_free(aux);
 }
#endif

  cur = xmlNewChild(doc->children, NULL, XML_MESSAGE_HEADER, NULL);

  doc_aux = node_to_xml(msg->header->from);
  aux = xmlNewChild(cur, NULL, XML_MESSAGE_HEADER_FROM, NULL);
  xmlAddChild(aux, xmlCopyNode(doc_aux->children, 1));
  xmlFreeDoc(doc_aux);
  
  aux = xmlNewChild(cur, NULL, XML_MESSAGE_HEADER_TO, NULL);
  length = util_num_length(msg->header->rcpt->type) + 1;
  type = g_new0(xmlChar, length);
  g_snprintf(type, length, MSG_RCPT_TYPE_FORMAT, msg->header->rcpt->type);
  xmlNewChild(aux, NULL, XML_MESSAGE_HEADER_TO_TYPE, type); 
  g_free(type);

  if (msg->header->rcpt->type == MSG_RCPT_NODE) {
    doc_aux = node_to_xml(msg->header->rcpt->to.node);
  } else if (msg->header->rcpt->type == MSG_RCPT_GROUP) {
    doc_aux = group_to_xml(msg->header->rcpt->to.group);
  }
  xmlAddChild(aux, xmlCopyNode(doc_aux->children, 1));
  xmlFreeDoc(doc_aux);

  length = util_num_length(msg->header->type) + 1;
  type = g_new0(xmlChar, length);
  g_snprintf(type, length, MSG_TYPE_FORMAT, msg->header->type);
  xmlNewChild(cur, NULL, XML_MESSAGE_HEADER_TYPE, type);
  g_free(type);

  cur = xmlNewChild(doc->children, NULL, XML_MESSAGE_DATA, NULL);
  xmlAddChild(cur, xmlCopyNodeList(msg->data->info->children));

  //  log_debug("[msg to xml] exit");

  return doc;
}


/**
 * Creates a message from an XML structure.
 * \param doc The document to be transformed into a message.
 * \return A new message created from the document. This message should be destroyed when no longer needed.
 */
msg_t* msg_from_xml(const xmlDocPtr doc)
{
  msg_t* msg;

  //  log_debug("[msg from xml] init");

  if (strcmp(doc->children->name, XML_MESSAGE)) {
    /* not a message xml!!! */
    msg = NULL;
  } else {
    xmlNodePtr cur;
    xmlDocPtr doc_aux;
    xmlChar *aux;
    group_t *group;
    node_t *node;

    doc_aux = xmlNewDoc(XML_VERS);
    
    msg = g_new(msg_t, 1);
    msg->header = g_new0(msg_header_t, 1);
    msg->data = g_new0(msg_data_t, 1);

    /* id */
#if USE_MSG_ID
    cur = doc->children;
    aux = xmlGetProp(cur, XML_NODE_ID);
    msg->id = g_strdup(aux);
    xmlFree(aux);
#endif

    /* header */
    cur = doc->children->children->children;
    doc_aux->children = xmlCopyNode(cur->children, 1);
    node = node_from_xml(doc_aux);
    g_assert(node != NULL);
    xmlFreeNodeList(doc_aux->children);
    msg->header->from = node;
    
    cur = cur->next;
    msg->header->rcpt = g_new0(msg_rcpt_t, 1);
    aux = xmlNodeGetContent(cur->children);
    msg->header->rcpt->type = atoi(aux);
    xmlFree(aux);

    doc_aux->children = xmlCopyNode(cur->children->next, 1);
    group = group_from_xml(doc_aux);
    if (group != NULL) {
      msg->header->rcpt->to.group = group;
      //    msg->header->rcpt->type = MSG_RCPT_GROUP;
    } else {
      node = node_from_xml(doc_aux);
      g_assert(node != NULL);
      msg->header->rcpt->to.node = node;
      //    msg->header->rcpt->type = MSG_RCPT_NODE;
    }
    xmlFreeNodeList(doc_aux->children);
    
    cur = cur->next;
    aux = xmlNodeGetContent(cur);
    msg->header->type = atoi(aux);
    xmlFree(aux);
    
    /* data */
    cur = doc->children->children->next;
    doc_aux->children = xmlCopyNodeList(cur->children);
    msg->data->info = doc_aux;
  }

  //  log_debug("[membership from xml] exit");

  return msg;
}
