/* GMSD - Group Membership Service Daemon
 * Copyright (C) 2003 Nelio Pereira (nelio@ime.usp.br)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

#include <string.h>

//#include <glib.h>
//#include <libxml/tree.h>

#include "log.h"
#include "node.h"
#include "util.h"
#include "xml.h"



/**
 * Creates a new node.
 * \return A new node.
 */
node_t* node_new(const node_id_t id, gchar* const hostname, guint port) 
{
  node_t *node = g_new0(node_t, 1);
  node->id = id;
  node->hostname = g_strdup(hostname);
  node->port = port;
  return node;
}


/**
 *
 * \param node
 */
void node_destroy(node_t* node)
{
  //  log_debug("[node destroy] init");

  g_assert(node != NULL);
  g_free(node->hostname);
  g_free(node);

  //  log_debug("[node destroy] exit");
}


/**
 * Compares two nodes.
 * \param node1
 * \param node2
 * \return 
 */
gint node_compare(const node_t* const node1, 
		  const node_t* const node2)
{
  return node_compare_full(node1, node2, NULL);
}


/**
 * Compares two nodes.
 * \param node1
 * \param node2
 * \param user_data
 * \return 
 */
gint node_compare_full(const node_t* const node1, 
		       const node_t* const node2, 
		       gpointer user_data)
{
  gint r;

  g_assert(node1 != NULL);
  g_assert(node2 != NULL);

  return (r = strcmp(node1->hostname, node2->hostname)) ? r : (node1->port - node2->port);
}


/**
 * Compares two nodes by theis node identifiers.
 * \param node1
 * \param node2
 * \return 
 */
gint node_compare_by_id(const node_t* const node1, const node_t* const node2) 
{
  return node_id_compare(node1->id, node2->id);
}


/**
 * Compares two node identifiers.
 * \param nodeid1
 * \param nodeid2
 * \return 
 */
gint node_id_compare(const node_id_t nodeid1, const node_id_t nodeid2)
{
  return nodeid1 - nodeid2;
}


/**
 *
 * \param node
 * \return xmlDocPtr
 */
xmlDocPtr node_to_xml(const node_t* const node)
{
  xmlDocPtr doc;
  xmlChar *aux;
  guint length;

  //  log_debug("[node to xml] init");

  doc = xmlNewDoc(XML_VERS);
  doc->children = xmlNewDocRawNode(doc, NULL, XML_NODE, NULL);
  xmlNewChild(doc->children, NULL, XML_NODE_HOSTNAME, node->hostname);

  length = util_num_length(node->port) + 1;
  aux = g_new0(xmlChar, length);
  g_snprintf(aux, length, NODE_PORT_FORMAT, node->port);
  xmlNewChild(doc->children, NULL, XML_NODE_PORT, aux);
  g_free(aux);

  length = util_num_length(node->id) + 1;
  aux = g_new0(xmlChar, length);
  g_snprintf(aux, length, NODE_ID_FORMAT, node->id);
  xmlSetProp(doc->children, XML_NODE_ID, aux);
  g_free(aux);

  //  log_debug("[node to xml] exit");

  return doc;
}


/**
 *
 * \param doc
 * \return
 */
node_t* node_from_xml(const xmlDocPtr doc)
{
  node_t* node;

  //  log_debug("[node from xml] init");

  if (strcmp(doc->children->name, XML_NODE)) {
    /* not a node xml!!! */
    node = NULL;
  } else {
    node_id_t node_id;
    gchar *hostname;
    guint port;
    xmlNodePtr cur;
    xmlChar *aux;

    cur = doc->children;
    aux = xmlGetProp(cur, XML_NODE_ID);
    node_id = atoi(aux);
    xmlFree(aux);
    
    cur = cur->children;
    hostname = xmlNodeGetContent(cur);
    
    cur = cur->next;
    aux = xmlNodeGetContent(cur);
    port = atoi(aux);
    xmlFree(aux);

    node = node_new(node_id, hostname, port);
    xmlFree(hostname);
  }

  //  log_debug("[node from xml] exit");

  return node;
}
