/* GMSD - Group Membership Service Daemon
 * Copyright (C) 2003 Nelio Pereira (nelio@ime.usp.br)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

/*! compilation definitions */
#include "config.h"


//#include <time.h>

//#include <glib.h>
//#include <libxml/tree.h>

//#include "group.h"
//#include "group_id.h"
#include "ipc.h" 
#include "log.h"
//#include "msg.h" 
//#include "membership.h"
#include "protocol.h"
#include "timer.h"
#include "util.h"
#include "xml.h"


/*! The detection method \em circular-token. See \ref DETECTION_METHOD. */
#define CIRCULAR_TOKEN 1

/*! The detection method \em PMC. See \ref DETECTION_METHOD. */
#define PMC 2


/**
 * The method for detecting failures. It may be:
 * - CIRCULAR_TOKEN: a token will rotate through the logical ring.
 * - PMC: each node checks its right neighbor.
 */
#define DETECTION_METHOD CIRCULAR_TOKEN


#if THREADS_ENABLED

/*! Macro used to acquire the context of the default main loop. */  
#define ACQUIRE_CONTEXT  g_main_context_wait(NULL, cond, mutex)
/*! Macro used to release the context of the default main loop. */
#define RELEASE_CONTEXT  g_main_context_release(NULL)

#endif


/*! \mainpage Servio de Pertinncia 
 *
 * <center>The \em One-Round Processor Group Membership Algorithm, 
 * by Flaviu Cristian and Frank Schmuck</center>
 *
 * \section intro Introduo
 *
 * This is the introduction.
 *
 * \section install Installation
 *
 * \subsection step1 Step 1: Opening the box
 *  
 * etc...
 */


/* * * *  Protocol Variables * * * */

/*! The \em MissingNeighbor timer. */
static mytimer_t *neighbor_timer;

/*! The \em Send-I-Am-Alive timer. */
static mytimer_t *alive_timer;

/*! The \em Send-Probe. */
static mytimer_t *probe_timer; 

/*! This node. */
static node_t *mynode;

/*! The neighbor timeout. */
static guint neighbor_period;

/*! The probe timeout. */
static guint probe_period;

/*! The delay timeout. */
static guint timeout_delay;

#if THREADS_ENABLED

/*! */
static GMutex *mutex;

/*! */
static GCond *cond; 

#endif

/* * * * * * * * * * * * * * * * * */


static void protocol_neighbor_timer(gpointer);
static void protocol_alive_timer(gpointer);
static void protocol_probe_timer(gpointer);

static void protocol_init_timestamps(const node_t* const, GHashTable* const);

static void protocol_new_group_info(const membership_t* const);
static void protocol_create_new_group(const group_id_t* const);
static void protocol_join(group_id_t* const);

static void protocol_send_alive(node_t* const, const group_id_t* const, const time_t);
static void protocol_send_membership(msg_rcpt_t* const, const msg_type_t, const membership_t* const);

static void protocol_receive_new_group(const node_t* const, const membership_t* const);
static void protocol_receive_alive(const node_t* const, const group_id_t* const, const time_t);
static void protocol_receive_probe(const node_t* const, const membership_t*);


/*! The mainloop. */
extern GMainLoop    *mainloop;
/*! The membership list of all possible nodes in the group. */
extern group_t      *membership_list;
/*! The current membership. */
extern membership_t *membership;


/*! Internal variable used to identify if this module has already been initialized */
static gboolean enabled = FALSE;



/**
 * Initializes the protocol routines and data structures.
 * \param gid_filename The file name where the group identifier should be persisted.
 * \param _neighbor_period The neighbor period.
 * \param _probe_period The probe period.
 * \param _timeout_delay The timeout delay.
 */
void protocol_init(const gchar* const gid_filename,
		   const guint _neighbor_period,
		   const guint _probe_period,
		   const guint _timeout_delay)
{
  log_debug("[protocol init] init");

#if THREADS_ENABLED
  if (! g_thread_supported()) { 
    g_thread_init(NULL);
  }
  mutex = g_mutex_new();
  cond = g_cond_new();
#endif

  /* initialize this node */
  mynode = membership_get_mynode();
  g_assert(mynode != NULL);
  
  /* initialize the membership */
  g_assert(membership_list != NULL);
  g_assert(membership != NULL);

  /* initialize the current gid */
  group_id_persist_init(membership->gid, gid_filename);
  group_id_set(membership->gid, membership->gid->seq + 1, membership->gid->node_id);
  group_id_store(membership->gid);

  /* add this node to the membership */ 
  group_add_node(membership->group, mynode);

  /* initialize the timestamps */
  g_list_foreach(membership_list->members, 
		 (GFunc)protocol_init_timestamps, 
		 membership->timestamps);
  g_hash_table_insert(membership->timestamps, 
		      GINT_TO_POINTER(mynode->id), 
		      GINT_TO_POINTER(time(NULL)));

  /* inform local clients about the initial membership */
  ipc_membership_changed(membership);

  /* initialize the timers */
  neighbor_timer = timer_new(protocol_neighbor_timer, neighbor_timer);
  alive_timer    = timer_new(protocol_alive_timer, alive_timer);
  probe_timer    = timer_new(protocol_probe_timer, probe_timer);
  neighbor_period = _neighbor_period;
  probe_period    = _probe_period;
  timeout_delay   = _timeout_delay;

  /* initialize the messages handling routines */
  msg_add_handle(PROTOCOL_MSG_NEW_GROUP, protocol_handle_msg_new_group);
  msg_add_handle(PROTOCOL_MSG_ALIVE, protocol_handle_msg_alive);
  msg_add_handle(PROTOCOL_MSG_PROBE, protocol_handle_msg_probe);

  timer_set(probe_timer, 0);

  enabled = TRUE;
  log_debug("[protocol init] exit");
}


/**
 * Routine called to initialize a timestamps hash table.
 * \param node The node to be initialized in the hash table.
 * \param timestamps The timestamps hash table to be updated.
 */
static void protocol_init_timestamps(const node_t* const node, GHashTable* const timestamps)
{
  static gpointer zero = GINT_TO_POINTER(0);
  g_hash_table_insert(timestamps, GINT_TO_POINTER(node->id), zero);
} 


/**
 * Destroys the protocol routines and data structures.
 */
void protocol_finalize()
{
#if THREADS_ENABLED
  ACQUIRE_CONTEXT;
#endif

  log_debug("[protocol finalize] init");

  g_mutex_free(mutex);
  g_cond_free(cond);

  timer_destroy(neighbor_timer);
  timer_destroy(alive_timer);
  timer_destroy(probe_timer);

  group_id_persist_destroy();

  enabled = FALSE;

#if THREADS_ENABLED
  RELEASE_CONTEXT;
#endif

  log_debug("[protocol finalize] exit");
}


/**
 * Informs if this modules has been initialized.
 * \return A boolean that indicates if this modules has been initialized.
 */
gboolean protocol_enabled()
{
  return enabled;
}


/**
 * Handles the \em Missing-Neighbor timer.
 * \param data (unused)
 */
static void protocol_neighbor_timer(gpointer data) 
{
  /* "I am alive" message from left neighbor is missing */

  node_t* left;

#if THREADS_ENABLED
  ACQUIRE_CONTEXT;
#endif

  log_debug("[neighbor timer] init");
  log_notice("neighbor timer - missing I-Am-Alive message from left neighbor.") ;

  left = membership_left_neighbor(membership, mynode);
  group_remove_node(membership->group, left);
  protocol_create_new_group(membership->gid);

#if THREADS_ENABLED
  RELEASE_CONTEXT;
#endif

  log_debug("[neighbor timer] exit");
}
	
				  
/**
 * Handles the \em I-Am-Alive timer.
 * \param data (unused)
 */
static void protocol_alive_timer(gpointer data) 
{
  /* It's time to send "I am alive" message to right neighbor. */

  node_t *right;

#if THREADS_ENABLED
  ACQUIRE_CONTEXT;
#endif

  log_debug("[alive timer] init");
  log_notice("alive timer - time to send I-Am-Alive message to right neighbor.") ;

  right = membership_right_neighbor(membership, mynode);
  protocol_send_alive(right, membership->gid, time(NULL));

  log_notice("alive timer - waking up in %u milliseconds", neighbor_period);
  timer_set(alive_timer, neighbor_period);

  
#if THREADS_ENABLED
  RELEASE_CONTEXT;
#endif

  log_debug("[alive timer] exit");
}


/**
 * Handles the \em Probe timer routine.
 * Note that probes are sent only by the group leaders of minority 
 * groups because in procedure join, the probe timer is set 
 * only if this processor is the leader of a minority group.
 * \param data (unused)
 */
static void protocol_probe_timer(gpointer data) 
{
  /* It's time to send a probe message */
  GList *list;
  group_t *group;
  node_t *node;
  guint size;

#if THREADS_ENABLED
  ACQUIRE_CONTEXT;
#endif

  log_debug("[probe timer] init");
  log_notice("probe timer - time to look for new members.") ;

  g_hash_table_insert(membership->timestamps, 
		      GINT_TO_POINTER(mynode->id), 
		      GINT_TO_POINTER(time(NULL)));

  /* send message to P - members */
  group = group_new();
  for (list = membership_list->members; list != NULL; list = g_list_next(list)) {
    node = (node_t*)list->data;
    if (! group_in(membership->group, node)) {
      group_add_node(group, node);
    }
  }
  size = group_size(group);
  if (size > 0) {
    msg_rcpt_t rcpt;
    if (size == 1) {
      rcpt.to.node = (node_t*)group->members->data;
      rcpt.type = MSG_RCPT_NODE;
    } else {
      rcpt.to.group = group; 
      rcpt.type = MSG_RCPT_GROUP;
    }
    protocol_send_membership(&rcpt, PROTOCOL_MSG_PROBE, membership);
  }
  group_destroy(group);
  
  log_notice("probe timer - waking up in %u milliseconds", probe_period);
  timer_set(probe_timer, probe_period);

#if THREADS_ENABLED
  RELEASE_CONTEXT;
#endif

  log_debug("[probe timer] exit");
}


/**
 * Process a new group information, and verifies if a new group must be created.
 * \param new_membership The new membership to be verified.
 */
static void protocol_new_group_info(const membership_t* const new_membership)
{
  gboolean my_info_uptodate    = TRUE;
  gboolean other_info_uptodate = TRUE;
  node_t *p;
  gboolean p_in_members        = FALSE;
  gboolean p_in_new_members    = FALSE;
  GList *list;
  gint t1;
  gint t2;
  gint gid_compare;
  
  log_debug("[new group info] init");

  /* Merge my knowledge of the state of the system with the new information. */
  for (list = membership_list->members; list != NULL; list = g_list_next(list)) {
    p = (node_t*)list->data;
 
    p_in_members = group_in(membership->group, p);
    p_in_new_members = group_in(new_membership->group, p);

    t1 = GPOINTER_TO_INT(g_hash_table_lookup(membership->timestamps, 
					     GINT_TO_POINTER(p->id)));
    t2 = GPOINTER_TO_INT(g_hash_table_lookup(new_membership->timestamps, 
					     GINT_TO_POINTER(p->id)));

    if (! p_in_members &&  p_in_new_members) {
      if (t1 < t2) {
	group_add_node(membership->group, p);
	my_info_uptodate = FALSE;
      } else {
	other_info_uptodate = FALSE;
      } 
    } 
    else if (p_in_members && ! p_in_new_members) {
      if ((t1 <= t2) && (node_compare(mynode, p) != 0)) {
	group_remove_node(membership->group, p);
	my_info_uptodate = FALSE;
      } else {
	other_info_uptodate = FALSE;
      } 
    } 
    g_hash_table_insert(membership->timestamps, 
			GINT_TO_POINTER(p->id),
			GINT_TO_POINTER(MAX(t1, t2)));
  } /* for all p in P */
  
  gid_compare = group_id_compare(membership->gid, new_membership->gid);
  if (gid_compare < 0) {
    if (other_info_uptodate) {
      protocol_join(new_membership->gid);
    } else {
      protocol_create_new_group(new_membership->gid);
    }
  } else if (gid_compare > 0) {
    if (my_info_uptodate) {
      log_debug("[new group info] Membership already updated. Nothing changed.");
    } else {
      protocol_create_new_group(membership->gid);
    }
  }
  
  log_debug("[new group info] exit");
}


/**
 * Creates a new group from an old group identifier.
 * \param old_gid The old group identifier.
 */
static void protocol_create_new_group(const group_id_t* const old_gid)
{
  GList *list;
  node_t *node;
  group_t *group;
  group_id_t *new_gid;
  guint size;

  log_debug("[create new group] init");

  new_gid = group_id_new(old_gid->seq + 1, mynode->id);
  g_hash_table_insert(membership->timestamps, 
		      GINT_TO_POINTER(mynode->id), 
		      GINT_TO_POINTER(time(NULL)));

  /* send new-group to P - {myid} */
  group = group_new();
  for (list = membership_list->members; list != NULL; list = g_list_next(list)) {
    node = (node_t*)list->data;
    if (node_compare(node, mynode) != 0) {
      group_add_node(group, node);
    }
  }
  size = group_size(group);
  if (size > 0) {
    msg_rcpt_t rcpt;
    membership_t *new_membership;
    if (size == 1) {
      rcpt.to.node = (node_t*)group->members->data;
      rcpt.type = MSG_RCPT_NODE;
    } else {
      rcpt.to.group = group;
      rcpt.type = MSG_RCPT_GROUP;
    }
    new_membership = g_new0(membership_t, 1);
    new_membership->gid = new_gid;
    new_membership->group = membership->group;
    new_membership->timestamps = membership->timestamps;
    protocol_send_membership(&rcpt, PROTOCOL_MSG_NEW_GROUP, new_membership);
    g_free(new_membership);
  }
  group_destroy(group);
  
  protocol_join(new_gid);
  group_id_destroy(new_gid);

  log_debug("[create new group] exit");
}


/**
 * Joins to the membership described by a given group identifier.
 * \param new_gid The new group identifier.
 */
static void protocol_join(group_id_t* const new_gid)
{
  gboolean is_leader;

  log_debug("[join] init");

  /* update current gid */
  group_id_destroy(membership->gid);
  membership->gid = group_id_copy(new_gid);
  group_id_store(membership->gid);

  //  if (log_info_enabled()) {
  //    gchar* aux = group_to_string(membership->group);
  //    log_info("joined group (" GROUP_ID_SEQ_FORMAT "," NODE_ID_FORMAT "): {%s}", 
  //	     membership->gid->seq, membership->gid->node_id, aux);
  //    g_free(aux);
  //  }
  
  /* cancel all timers */
  timer_cancel(probe_timer);
  timer_cancel(alive_timer);
  timer_cancel(neighbor_timer);

  /* set new timers */
  is_leader = membership_is_leader(membership, mynode);
  if (is_leader && ! membership_majority(membership)) {
    log_notice("probe timer - waking up in %u milliseconds", probe_period);
    timer_set(probe_timer, probe_period);
  }
  if (group_size(membership->group) > 1) {
    guint timeout;

#if DETECTION_METHOD == CIRCULAR_TOKEN
    if (is_leader) {
#endif
      timeout = neighbor_period - timeout_delay;
      log_notice("alive timer - waking up in %u milliseconds", timeout);
      timer_set(alive_timer, timeout);
#if DETECTION_METHOD == CIRCULAR_TOKEN
    }
    timeout = neighbor_period + membership_rank(membership, mynode)*timeout_delay;
#elif DETECTION_METHOD == PMC
    timeout = neighbor_period + timeout_delay;
#endif    

    log_notice("neighbor timer - waking up in %u milliseconds", timeout);
    timer_set(neighbor_timer, timeout);
  }

  /* finally, let all local clients know about the change. */
  ipc_membership_changed(membership);

  log_debug("[join] exit");
}


/**
 * Routine called to process the content of a \em New-Group message.
 * \param from The sender node.
 * \param new_membership The membership received in the \em New-Group message.
 */
void protocol_receive_new_group(const node_t* const from, const membership_t* const new_membership)
{
  log_debug("[receive new_group] init");

  protocol_new_group_info(new_membership);

  log_debug("[receive new_group] exit");
}


/**
 * Routine called to process the content of a \em I-Am-Alive message.
 * \param from The sender node.
 * \param new_gid The group identifier received in the \em I-Am-Alive message.
 * \param timestamp The timestamp received in the \em I-Am-Alive message.
 */
void protocol_receive_alive(const node_t* const from, const group_id_t* const new_gid, const time_t timestamp)
{
  log_debug("[receive alive] init");

  g_assert(from != NULL);
  g_assert(new_gid != NULL);
  g_assert(timestamp != 0);

  if (group_id_compare(membership->gid, new_gid) == 0) {
    guint timeout;

    log_debug("recebendo alive de %d:%s:%d", from->id, from->hostname, from->port); 

    // timestamps(p) <- t'
    g_hash_table_insert(membership->timestamps, 
			GINT_TO_POINTER(from->id),
			GINT_TO_POINTER(timestamp));

#if DETECTION_METHOD == CIRCULAR_TOKEN 
    timeout = neighbor_period + membership_rank(membership, mynode)*timeout_delay;

    if (! membership_is_leader(membership, mynode)) {
      node_t *right = membership_right_neighbor(membership, mynode);
      protocol_send_alive(right, membership->gid, time(NULL));
    }
#elif DETECTION_METHOD == PMC
    timeout = neighbor_period + timeout_delay;
#endif

    log_notice("neighbor timer - waking up in %u milliseconds", timeout);
    timer_set(neighbor_timer, timeout);
    
  } /* if gid' == CurrentGid */

  log_debug("[receive alive] exit");
}


/**
 * Routine called to process the content of a \em Probe message.
 * \param from The sender node.
 * \param new_membership The membership received in the \em Probe message.
 */
void protocol_receive_probe(const node_t* const from, const membership_t* const new_membership)
{
  log_debug("[receive probe] init");
  
  protocol_new_group_info(new_membership);

  log_debug("[receive probe] exit");
}


/**
 * Sends a \em I-Am-Alive message to a given node.
 * \param to The recipient node.
 * \param group_id The group identifier to be sent in this message.
 * \param timestamp The timestamp to be sent in this message.
 */
static void protocol_send_alive(node_t* const to, const group_id_t* const group_id, const time_t timestamp)
{
  msg_t* msg;
  msg_rcpt_t rcpt;
  xmlDocPtr doc, doc_aux;
  xmlChar *aux;
  guint length;

  log_debug("[send alive] init");

  /** 
   * alive message data:
   *
   * <timestamp></timestamp>
   * <group_id sequence="" node_id=""/>
   */
  
  doc = xmlNewDoc(XML_VERS);

  length = util_num_length(timestamp) + 1;
  aux = g_new0(xmlChar, length);
  g_snprintf(aux, length, MEMBERSHIP_TIMESTAMP_FORMAT, timestamp);
  doc->children = xmlNewDocNode(doc, NULL, XML_TIMESTAMP, aux);
  g_free(aux);

  doc_aux = group_id_to_xml(group_id);
  xmlAddSibling(doc->children, xmlCopyNode(doc_aux->children, 1));
  xmlFreeDoc(doc_aux);

  rcpt.to.node = to;
  rcpt.type = MSG_RCPT_NODE;
  msg = msg_new(mynode, &rcpt, PROTOCOL_MSG_ALIVE, doc);
  xmlFreeDoc(doc);

#if 0
  if (log_debug_enabled()) {
    xmlDocPtr ptr = msg_to_xml(msg);
    log_debug("Alive Message:");
    xmlSaveFormatFile("-", ptr, 1);
    xmlFreeDoc(ptr);
  }
#endif

  msg_send(msg);
  msg_destroy(msg);
  
  log_debug("[send alive] exit");
}


/**
 * Sends a message with a given membership to a recipient.
 * \param rcpt The recipient of the message.
 * \param msg_type The type of the message to be sent.
 * \param membership The membership to send in this message.
 */
static void protocol_send_membership(msg_rcpt_t* const rcpt, 
				     const msg_type_t msg_type, 
				     const membership_t* const membership)
{
  msg_t* msg;
  xmlDocPtr doc;

  log_debug("[send membership] init");

  /**
   * membership message data:
   *
   * <membership>
   *  ...
   * </membership>
   */

  doc = membership_to_xml(membership);
  msg = msg_new(mynode, rcpt, msg_type, doc);
  xmlFreeDoc(doc);

#if 0
  if (log_debug_enabled()) {
    xmlDocPtr ptr = msg_to_xml(msg);
    log_debug("Membership Message:");
    xmlSaveFormatFile("-", ptr, 1);
    xmlFreeDoc(ptr);
  }
#endif

  msg_send(msg);
  msg_destroy(msg);
  
  log_debug("[send membership] exit");
}


/**
 * Routine called whenever a \em New-Group message arrives.
 * \param msg The incoming \em New-Group message.
 */
void protocol_handle_msg_new_group(const msg_t* const msg)
{
  membership_t *memb;

#if THREADS_ENABLED
  ACQUIRE_CONTEXT;
#endif

  log_debug("[protocol handle msg new_group] init");

  memb = membership_from_xml(msg->data->info);
  g_assert(memb != NULL);
  protocol_receive_new_group(msg->header->from, memb);
  membership_destroy_full(memb);

#if THREADS_ENABLED
  RELEASE_CONTEXT;
#endif

  log_debug("[protocol handle msg new_group] exit");
}


/**
 * Routine called whenever a \em I-Am-Alive message arrives.
 * \param msg The incoming \em I-Am-Alive message.
 */
void protocol_handle_msg_alive(const msg_t* const msg)
{
  group_id_t* group_id;
  time_t timestamp;
  xmlDocPtr doc;
  xmlNodePtr cur;
  xmlChar *aux;

#if THREADS_ENABLED
  ACQUIRE_CONTEXT;
#endif

  log_debug("[protocol handle msg alive] init");

  cur = msg->data->info->children;
  aux = xmlNodeGetContent(cur);
  timestamp = atol(aux);
  g_assert(timestamp > 0);
  xmlFree(aux);

  cur = cur->next;
  doc = xmlNewDoc(XML_VERS);
  doc->children = xmlCopyNode(cur, 1);
  group_id = group_id_from_xml(doc);
  g_assert(group_id != NULL);
  xmlFreeDoc(doc);
  
  protocol_receive_alive(msg->header->from, group_id, timestamp);
  group_id_destroy(group_id);
  
#if THREADS_ENABLED
  RELEASE_CONTEXT;
#endif

  log_debug("[protocol handle msg alive] exit");
}


/**
 * Routine called whenever a \em Probe message arrives.
 * \param msg The incoming \em Probe message.
 */
void protocol_handle_msg_probe(const msg_t* const msg)
{
  membership_t *memb;
  
#if THREADS_ENABLED
  ACQUIRE_CONTEXT;
#endif

  log_debug("[protocol handle msg probe] init");

  memb = membership_from_xml(msg->data->info);
  g_assert(memb != NULL);
  protocol_receive_probe(msg->header->from, memb);
  membership_destroy_full(memb);

#if THREADS_ENABLED
  RELEASE_CONTEXT;
#endif

  log_debug("[protocol handle msg probe] exit");
}
