/* GMSD - Group Membership Service Daemon
 * Copyright (C) 2003 Nelio Pereira (nelio@ime.usp.br)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

#include <stdio.h>

//#include <glib.h>

#include "timer.h"


static gboolean timer_execute(mytimer_t*);


/**
 * Creates a new \ref mytimer_t.
 * \param func
 * \param func_data
 * \return 
 */
mytimer_t* timer_new(const timer_func_t func, const gpointer func_data)
{
  mytimer_t* timer = g_new0(mytimer_t, 1);
  timer->id = TIMER_ID_DEFAULT;
  timer->func = func;
  timer->func_data = func_data;
  timer->executing = FALSE;
  return timer;
}


/**
 * Destroys a timer.
 * \param timer
 */
void timer_destroy(mytimer_t* const timer)
{
  if (timer_enabled(timer)) {
    timer_cancel(timer);
  }
  g_free(timer);
}


/**
 * Verifies if a given timer is enabled.
 * \param timer The timer to be tested.
 * \return A boolean indicating the timer status.
 */
gboolean timer_enabled(const mytimer_t* const timer)
{
  return (timer->id != TIMER_ID_DEFAULT);
}


/**
 * Sets a timer with a given timeout.
 * \param timer The timer to be set.
 * \param timeout The timeout to set the timer.
 */
void timer_set(mytimer_t* const timer, const guint timeout)
{
  /**
   * if the timer is not enabled or if it's running, there's no need to 
   * cancel it before setting it. If the timer is running, it'll be 
   * automatically stopped afterwards.
   */
  if (timer_enabled(timer) && ! timer->executing) {
    timer_cancel(timer);
  }
  timer->timeout = timeout;
  timer->id = g_timeout_add(timer->timeout, (GSourceFunc)timer_execute, timer);
  /*
  timer->id = g_timeout_add_full(G_PRIORITY_HIGH_IDLE,
				 timer->timeout,
				 (GSourceFunc)timer_execute,
				 timer,
				 NULL);
  */
}


/**
 * Cancels a given timer.
 * \param timer The timer to be cancelled.
 */
void timer_cancel(mytimer_t* const timer)
{
  gboolean r;

  g_assert(timer != NULL);

  if (! timer_enabled(timer)) {
     /* nothing to be done. */
    return;
  }

  r = g_source_remove(timer->id);
  g_assert(r == TRUE);

  timer->id = TIMER_ID_DEFAULT;
}


/**
 * Internal routine that is called whenever a timer is triggered.
 * We assume that once the timer is triggered, the function will be called just once.
 * If it's needed to schedule the timer again, the application should call timer_set
 * afterwards.
 * \param timer The timer being executed.
 * \return A boolean (in this case, always FALSE) indicating that this timer should be destroyed.
 */
static gboolean timer_execute(mytimer_t* timer)
{
  timer->executing = TRUE;
  timer->func(timer->func_data);
  timer->executing = FALSE;
  return FALSE;
}
