use Time::ParseDate;
use strict;
use sort 'stable';



sub mylog
{
    my ($line) = @_;
    my $ctime = localtime;
    print STDERR "[ $ctime ] " . $line . "\n";
}


##
# Parses a log to create the data structure. 
#
# param  filename - The log file to be parsed.
# return The events described in the log.
##
sub parse_log
{
    my ($filename) = @_;
    my ($events, $fp);

    open $fp, $filename || die $!;
    $events = process_file($fp);
    close $fp;
    
    return $events;
}



##
# Creates a data scruture from a given file pointer.
#
# param  fp - The file pointer to be used.
# return The data structure describing the log.
##
sub process_file
{
    my ($fp) = @_;
    my ($line, %events, $event);
    
    foreach $line (<$fp>) {
	$event = process_line($line);
	$events{$event->{'timestamp'}} = $event;
    }
    
    return \%events;
}


##
# Process a line to create its representation.
#
# param  line - The line to be processed.
# return The data structure representing that line.
##
sub process_line
{
    my ($line) = @_;
    my ($msg, $date, $msec, $host);
    
    if ($line !~ /^\[(.{24}) - (\d{6})\] gmsd ([^ ]+) [^:]+: ([^\n\r]+)/) {
	next;
    }
    
    $date = $1;
    $msec = $2;
    $host = $3;
    $msg  = $4;

    my %event = ();
    if ($msg eq 'server up. waiting for events.') {
#    if ($msg eq 'starting gmsd server...') {
	$event{'type'} = 'INIT';
    } elsif ($msg eq 'server down. releasing resources.') {
	$event{'type'} = 'DOWN';
    } elsif ($msg =~ /^message ([^ ]+) received \(type (\d+), recipient type (\d+)\)\.$/) {
	$event{'type'}          = 'MESSAGE_RECEIVED';
	$event{'msg_id'}        = $1;
	$event{'msg_type'}      = $2;
	$event{'msg_rcpt_type'} = $3;
	$event{'msg_from'}      = substr $1, 0, index($1,'.');
#    } elsif ($msg =~ /^message ([^ ]+) sent \(type (\d+), recipient type (\d+)\)\.$/) {
    } elsif ($msg =~ /^sending message ([^ ]+) \(type (\d+), recipient type \d+\).$/) {
	$event{'type'}          = 'MESSAGE_SENT';
	$event{'msg_id'}        = $1;
	$event{'msg_type'}      = $2;
	$event{'msg_rcpt_type'} = $3;
	$event{'msg_from'}      = substr $1, 0, index($1,'.');
    } elsif ($msg =~ /^joined group \(([^\)]+)\): \{([^\}]+)\}( \(message (.+)\))?$/) {
	$event{'type'}    = 'JOIN';
	$event{'gid'}     = $1;
	$event{'members'} = $2;
	$event{'msg_id'}  = $4;
    } else {
#	print "!!$msg!!\n";
	next;
    }
    
    # timestamp with microseconds
    my ($timestamp) = parsedate($date, UK => 1) . $msec;
    
    $event{'timestamp'} = $timestamp;
    $event{'hostname'}  = $host; #sprintf "%.3d", $host;
    
    return \%event;
}


##
# Creates a hastable describing each moment when there is a state change. Each of these events
# has elements describing what nodes are up/down, besides the detail of the event.
#
# param  events - The data structure representing the events
# return A hashtable describing the state changes
##
sub consolidate_events
{
    my ($events) = @_;
    my (%up, %nodes, %ret, @records, $timestamp, $event, %messages);

    my (%delta, $new_timestamp);
    
    foreach $timestamp (sort keys %$events) 
    {
	$event = $events->{$timestamp};
	my ($node) = $event->{'hostname'};
	$new_timestamp = 0;
	my (%record);
	
	if ($event->{'type'} eq 'INIT') {
	    $up{"$node"} = 1;
	    %delta = ();
	} elsif ($event->{'type'} eq 'DOWN') {
	    delete $up{"$node"};
	    delete $nodes{"$node"};
	} elsif ($event->{'type'} eq 'JOIN') {
	    # um n ir se unir a um grupo somente 
	    # depois de ter inicializado
#	    next if (! $up{"$node"});
	    
	    my (%aux);
	    $aux{'gid'} = $event->{'gid'};
	    $aux{'members'} = $event->{'members'};
	    $record{'msg_id'} = $event->{'msg_id'};
	    $nodes{"$node"} = \%aux;
	} elsif ($event->{'type'} eq 'MESSAGE_SENT') {
	    $record{'msg_id'}   = $event->{'msg_id'};
	    $record{'msg_from'} = $event->{'msg_from'};
	    $messages{$event->{'msg_id'}} = \%record if ($event->{'msg_id'} ne '(null)');
	} elsif ($event->{'type'} eq 'MESSAGE_RECEIVED') {
#	    if ( (defined $messages{$event->{'msg_id'}}) &&
#		 ($event->{'msg_from'} != $node) ) {
#		$delta{"$node"} += $timestamp - $messages{$event->{'msg_id'}} - 1;
#	    }
	    $record{'msg_id'} = $event->{'msg_id'};
	} else {
	    # the other events don't matter
	    next;
	}

#	$new_timestamp = ($timestamp - $delta{"$node"});
	$new_timestamp = $timestamp;
	$record{'timestamp'} = $new_timestamp;
	$record{'type'}      = $event->{'type'};
	$record{'hostname'}  = $node;

	my (%aux) = %up;
	$record{'up'} = \%aux;
	
	my (%aux1) = %nodes;
	$record{'nodes'} = \%aux1;
	
	push @records, \%record;
    }

    $ret{'records'} = \@records;
    $ret{'messages'} = \%messages;

    return \%ret;
}


####
# Knows how to sort event records.
# 
##
sub sort_records
{
    $a->{'timestamp'} <=> $b->{'timestamp'};
}


##
# Prints the events generated by consolidate_events
#
# param  events - The data structure representing the events
##
sub print_events
{
    my ($events) = @_;
    my ($key, $up, $nodes, $timestamp, $event);

    foreach $timestamp (sort {$a <=> $b} keys %$events) {
	$event = $events->{$timestamp};

	# first prints out the timestamp
	print $event->{'timestamp'} . " - ";
	print $event->{'type'} . " - ";
	print $event->{'hostname'} . "\n";
	
	# then prints the UP vector
	$up = $event->{'up'};
	print "\t- nodes up:\t";
	
	foreach $key (sort {$a <=> $b} keys %$up) {
	    print $key . " ";
	}
	print "\n";
	
	# now, for each node prints out the gid and members
	$nodes = $event->{'nodes'};
	print "\t- nodes:\n";
	foreach $key (sort {$a <=> $b} keys %$nodes) {
	    my $aux = $nodes->{$key};
	    print "\t\t" . $key . "  -  ";
	    print "gid:   " . $aux->{'gid'} . "\t";
	    print "members:   " . $aux->{'members'} . "\n";
	}

	print "\n";
    }
}




1;
